''' Process Engineering Elements '''

import math
from typing import Literal, Sequence

from ..util import linspace
from ..segments import Segment, SegmentCircle, SegmentText, SegmentPoly, SegmentArc
from ..elements import Element
from ..types import XY


#
# 1. VESSELS AND TANKS
#

class Tank(Element):
    '''1.1 Tank, vessel
        
        Args:
            h: height
            w: width
        
        Anchors:
            * W
            * S
    '''
    
    def __init__(self, h: float = 12, w: float = 9, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0,0), (0, h), (w, h), (w, 0)]))
        
        self.anchors['W'] = (0, h*10/12)
        self.anchors['S'] = (w/2, 0)

        
class Cistern(Element):
    '''1.2 Container, Tank, Cistern
    
        Args:
            h: Height of the tank
            w: Width of the tank
            level: Level of the liquid in the tank (0..1)
       
       Anchors:
           * NW
           * N
           * NE
           * W
           * E
           * SW
           * S
           * SE
    '''
    
    def __init__(self,
                 h: float = 6,
                 w: float = 8,
                 level:float = 0.75,
                 *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(w, h), (w, 0), (0, 0), (0, h)]))
        self.segments.append(Segment([(w, h*level), (0, h*level)], lw=0.5))
        
        self.anchors['NW'] = (0.25*w, 1*h)
        self.anchors['N'] = (0.5*w, 1*h)
        self.anchors['NE'] = (0.75*w, 1*h)
        self.anchors['W'] = (0, 0.75*h)
        self.anchors['E'] = (1, 0.75*h)
        self.anchors['SW'] = (0, 0)
        self.anchors['S'] = (0.5*w, 0)
        self.anchors['SE'] = (1*w, 0) 

#
# 2. COLUMNS WITH INTERNALS
#

#
# 3. HEAT EXCHANGERS
#

class HeatExchanger(Element):
    '''3.1 Heat Exchanger (general), condenser
    
        Args:
            r: radius
    
        Anchors:
            * N
            * S
            * W
            * E
    '''
    def __init__(self, r: float = 4, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentCircle([0, 0], r))
        self.segments.append(Segment([(-r, 0), (-2/4*r, 0), (0, 2/4*r), (0, -2/4*r), (2/4*r, 0), (r, 0)]))
        
        self.anchors['N'] = (0, r)
        self.anchors['S'] = (0, -r)
        self.anchors['W'] = (-r, 0)
        self.anchors['E'] = (r, 0)
 
        
class PlateExchanger(Element):
    '''3.6 Heat Exchanger of plate type
    
        Args:
            h: Heigth
            w: Width
            n: Num. of plates (default 3)
            
        Anchors:
            * NW
            * SW
            * NE
            * SE
        '''
        
    def __init__(self, h: float = 4, w: float = 13, n: int = 3, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0,0), (0, h), (w, h), (w, 0)]))
        self.segments.append(Segment([(1/13*w, h), (12/13*w, 0)]))
        self.segments.append(Segment([(1/13*w, 0), (12/13*w, h)]))
# Diseño de draw.io        
#         self.segments.append(Segment([(-0.15*w, 0.1*h), (0, 0.1*h), (w, 0.9*h), (w*1.15, 0.9*h)]))
#         self.segments.append(Segment([(-0.15*w, 0.9*h), (0, 0.9*h), (w, 0.1*h), (w*1.15, 0.1*h)]))
        
        num_plates = n
        step = w/(num_plates+1)
        for i in range(num_plates):
            self.segments.append(Segment([((i+1)*step, 0), ((i+1)*step, h)]))

# Diseño de draw.io    
#         self.segments.append(Segment([(w*1.15, 0.1*h-0.075*h), (w*1.15, 0.1*h+0.075*h)]))
#         self.segments.append(Segment([(w*1.15, 0.9*h-0.075*h), (w*1.15, 0.9*h+0.075*h)]))
#         self.segments.append(Segment([(-w*0.15, 0.1*h-0.075*h), (-w*0.15, 0.1*h+0.075*h)]))
#         self.segments.append(Segment([(-w*0.15, 0.9*h-0.075*h), (-w*0.15, 0.9*h+0.075*h)]))
        
        self.anchors['NE'] = (12/13*w, h)
        self.anchors['NW'] = (1/13*w, h)
        self.anchors['SE'] = (1/13*w, 0)
        self.anchors['SW'] = (12/13*w, 0)

#
# 4. STEAM GENERATORS, FURNACES, RECOOLING DEVICE
#

class Boiler(Element):
    '''4.1 Boiler with dome
    
        Args:
            h: Heigth
            w: Width
            
        Anchors:
            * N
            * W
        '''
        
    def __init__(self, h: float = 10, w: float = 8, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0, 0), (w, 0), (w, h*8/10), (0, h*8/10)]))
        self.segments.append(SegmentArc(center = (w/2, h*8/10), width=4/8*w, height=4/10*h, theta1=0, theta2=180))
        
        self.anchors['N'] = (w/2, h*10/10)
        self.anchors['W'] = (0, h*6/8)

#
# 5. COOLING TOWER
#

class CoolingTower(Element):
    '''5.1 Cooling tower (general)
    
        Args:
            h: Heigth
            w: Width
            
        Anchors:
            * E
            * W
        '''
        
    def __init__(self, h: float = 10, w: float = 8, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0, 0), (w, 0), (w, h*2/10), (0, h*2/10)]))
        self.segments.append(SegmentPoly([(0, h*2/10), (w, h*2/10), (w*6/8, h), (w*2/10, h)]))
        
        self.anchors['E'] = (w, h*2/10)
        self.anchors['W'] = (0, h*2/8)


#
# 6. FILTERS, LIQUID FILTERS, GAS FILTERS
#

class Filter(Element):
    '''6.1 Liquid filter (general)
    
        Args:
            h: Heigth
            w: Width
            
        Anchors:
            * N
            * S
        '''
        
    def __init__(self, h: float = 10, w: float = 6, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0, 0), (w, 0), (w, h), (0, h)]))
        self.segments.append(Segment([(0, h*5/10), (w*1/6, h*5/10)]))
        self.segments.append(Segment([(w*2/6, h*5/10), (w*4/6, h*5/10)]))
        self.segments.append(Segment([(w*5/6, h*5/10), (w, h*5/10)]))
        
        self.anchors['N'] = (w/2, h)
        self.anchors['S'] = (w/2, 0)


#
# 7. SCREENING DEVICES, SIEVES, AND RAKES
#

class Sieve(Element):
    '''7.1 Screening device, sieve, strainer, general
        
        Args:
            h: Heigth
            w: Width
            
        Anchors:
            * N
            * S
            * E
        '''
        
    def __init__(self, h: float = 10, w: float = 6, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0, h), (w, h), (w, h*3/10), (w/2, 0), (0, h*3/10)]))
        self.segments.append(Segment([(0, h), (w, h*3/10)], ls = '--'))
        
        self.anchors['N'] = (w/2, h)
        self.anchors['S'] = (w/2, 0)
        self.anchors['E'] = (w, h*4/1)

#
# 8. SEPARATORS
#

#
# 9. CENTRIFUGES
#
        
class DiscCentrifuge(Element):
    '''9.4 Centrifuge, separator disc-type
        
        Args:
            h: height
            w: width
        
        Anchors:
            * N
            * NE
            * SE
    '''

    def __init__(self, h: float = 8, w: float = 8, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0,0), (0, h), (w, h), (w, 0)]))
        self.segments.append(Segment([(w/2, h*7/8), (w/2, -h*1/8)]))
        self.segments.append(Segment([(w*1/8, h*1/8), (w*7/8, h*1/8)]))
        self.segments.append(Segment([(w*1/8, h*3.5/8), (w/2, h*5/8), (w*7/8, h*3.5/8)]))
        self.segments.append(Segment([(w*1/8, h*5.5/8), (w/2, h*7/8), (w*7/8, h*5.5/8)]))
        
        self.anchors['N'] = (w/2, h)
        self.anchors['NE'] = (w, h*6/8)
        self.anchors['SE'] = (w, 0)

#
# 10. DRIER
#

#
# 11. CRUSHING/GRINDING MACHINES
#

#
# 12. MIXERS/KNEADERS
#

#
# 13. SHAPING MACHINES -- PROCESSING IN VERTICAL DIRECTION
#

#
# 14. SHAPING MACHINES -- PROCESING IN HORIZONTAL DIRECTION
#

#
# 15. LIQUID PUMPS
#

class Pump(Element):
    '''15.1 Pump, liquid type (general)
    
        Args:
            r: radius
    
        Anchors:
            * W
            * E
    '''
    def __init__(self, r: float = 3, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentCircle([0, 0], r))
        self.segments.append(Segment([(0, r), (r, 0)]))
        self.segments.append(Segment([(0, -r), (r, 0)]))
        
        self.anchors['W'] = (-r, 0)
        self.anchors['E'] = (r, 0)

#
# 16. COMPRESSORS, VACUUM PUMPS
#

#
# 17. BLOWERS, FANS
#

#
# 18. LIFTING, CONVEYING AND TRANSPORT EQUIPMENT
#

#
# 19. PROPORTIONERS, FEEDERS AND DISTRIBUTION FACILITIES
#

#
# 20. ENGINES
#

# 
# 21. VALVES
#

class Valve(Element):
    '''21.1 Valve (general)
        
        Anchors:
            * W
            * E
            * center
    '''
    
    def __init__(self, h: float = 2, w: float = 4, *d, **kwargs):
        super().__init__(*d, **kwargs)
        
        self.segments.append(SegmentPoly([(0,0), (w, h), (w, 0), (0, h)]))
        
        self.anchors['W'] = (0, 1/4*h)
        self.anchors['E'] = (w, 1/4*h)

#
# 22. CHECK VALVES
#

#
# 23. VALVES AND FITTINGS WITH SAFETY FUNCTION
#

#
# 24. FITTINGS
#

#
# 25. GRAPHYCAL SYMBOLS FOR PIPING
#

#
# 26. APPARATUS ELEMENTS
#

#
# 27. INTERNALS
        
#
# 28. AGITATORS, STIRRERS
#

class Stirrer(Element):
    '''28.1 Agitator (general), Stirrer (general)
        
        Args:
            l: Length of the axis
        
        Anchors:
            * N
    '''
    
    def __init__(self, h: float = 6, w: float = 4, l: float = 1, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(2/4*w, h*1/6), (2/4*w, h*l)]))
        self.segments.append(Segment([(0, 0), (0, 2/6*h), (w, 0), (w, 2/6*h)]))

#
# 29. INTERNAL CHARACTERISTICS AND BUILT-IN-COMPONENTS
#


''' Test for using schemdraw in a script with interactive pyplot window  '''
import schemdraw
import schemdraw.elements as elm

#schemdraw.use('svg')

d = schemdraw.Drawing()
d.add(elm.Resistor().label('1K'))
d.add(elm.Capacitor().down())
d.save('cap.svg')

d2 = schemdraw.Drawing()
d2.add(elm.Diode().fill(True))

d2.save('res.svg')
d2.draw()
d.draw()



''' Two-terminal element definitions '''

from __future__ import annotations
import math
from typing import Sequence

from .elements import Element, Element2Term, gap
from ..util import Point, linspace
from ..segments import Segment, SegmentArrow, SegmentArc, SegmentText, SegmentCircle, SegmentPoly


resheight = 0.25      # Resistor height
reswidth = 1.0 / 6   # Full (inner) length of resistor is 1.0 data unit


class ResistorIEEE(Element2Term):
    ''' Resistor (IEEE/U.S. style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), (0.5*reswidth, resheight), (1.5*reswidth, -resheight),
             (2.5*reswidth, resheight), (3.5*reswidth, -resheight),
             (4.5*reswidth, resheight), (5.5*reswidth, -resheight), (6*reswidth, 0)]))


class ResistorIEC(Element2Term):
    ''' Resistor as box (IEC/European style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), (0, resheight), (reswidth*6, resheight),
             (reswidth*6, -resheight), (0, -resheight), (0, 0),
             gap, (reswidth*6, 0)]))



class ResistorVarIEEE(ResistorIEEE):
    ''' Variable resistor (U.S. style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        super().__init__(**kwargs)
        self.segments.append(SegmentArrow((1.5*reswidth, -resheight*2),
                                          (4.5*reswidth, reswidth*3.5),
                                          headwidth=.16, headlength=.2))


class ResistorVarIEC(ResistorIEC):
    ''' Variable resistor (European style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(SegmentArrow((1*reswidth, -resheight*2),
                                          (5*reswidth, reswidth*3.5),
                                          headwidth=.16, headlength=.2))


class Thermistor(ResistorIEC):
    ''' Thermistor '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(0, -resheight-.2), (.2, -resheight-.2), (1, resheight+.2)]))


class PhotoresistorIEEE(ResistorIEEE):
    ''' Photo-resistor (U.S. style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(SegmentArrow((.7, .75), (.4, .4),
                                          headwidth=.16, headlength=.2))
        self.segments.append(SegmentArrow((1, .75), (.7, .4),
                                          headwidth=.16, headlength=.2))


class PhotoresistorIEC(ResistorIEC):
    ''' Photo-resistor (European style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(SegmentArrow((.7, .75), (.4, .4),
                                          headwidth=.16, headlength=.2))
        self.segments.append(SegmentArrow((1, .75), (.7, .4),
                                          headwidth=.16, headlength=.2))


class Capacitor(Element2Term):
    ''' Capacitor (flat plates)

        Args:
            polar: Add polarity + sign
    '''
    def __init__(self, *d, polar: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        capgap = 0.18
        self.segments.append(Segment([(0, 0), gap, (0, resheight), (0, -resheight), gap,
                                     (capgap, resheight), (capgap, -resheight), gap,
                                     (capgap, 0)]))
        if polar:
            self.segments.append(SegmentText((-capgap*1.2, capgap), '+'))


class Capacitor2(Element2Term):
    ''' Capacitor (curved bottom plate)

        Args:
            polar: Add polarity + sign
    '''
    def __init__(self, *d, polar: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        capgap = 0.18
        self.segments.append(Segment([(0, 0), gap, (0, resheight),
                                     (0, -resheight), gap, (capgap, 0)]))
        self.segments.append(SegmentArc((capgap*1.5, 0), width=capgap*1.5,
                                        height=resheight*2.5, theta1=105, theta2=-105))

        if polar:
            self.segments.append(SegmentText((-capgap*1.2, capgap), '+'))


class CapacitorVar(Capacitor):
    ''' Variable capacitor '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(SegmentArrow((-2*reswidth, -resheight),
                                          (3*reswidth, reswidth*2),
                                          headwidth=.2, headlength=.2))


class CapacitorTrim(Capacitor):
    ''' Trim capacitor '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        capgap = 0.18
        # Line endpoints
        p1 = Point((-1.8*reswidth, -resheight))
        p2 = Point((1.8*reswidth+capgap, resheight))
        dx = p2.x - p1.x
        dy = p2.y - p1.y

        self.segments.append(Segment([p1, p2]))

        tlen = .14
        theta = math.atan2(dy, dx) + math.radians(90)
        t1 = Point((p2.x + tlen * math.cos(theta),
                    p2.y + tlen * math.sin(theta)))
        t2 = Point((p2.x + tlen * math.cos(theta-math.pi),
                    p2.y + tlen * math.sin(theta-math.pi)))
        self.segments.append(Segment([t1, t2]))


class Crystal(Element2Term):
    ''' Crystal oscillator '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        xgap = 0.2
        self.segments.append(Segment(
            [(0, 0), gap, (0, resheight), (0, -resheight), gap,
             (xgap/2, resheight), (xgap/2, -resheight), (xgap*1.5, -resheight),
             (xgap*1.5, resheight), (xgap/2, resheight), gap,
             (xgap*2, resheight), (xgap*2, -resheight), gap, (xgap*2, 0)]))


class Diode(Element2Term):
    ''' Diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(0, 0), gap, (resheight*1.4, resheight),
                                      (resheight*1.4, -resheight), gap, (resheight*1.4, 0)]))
        self.segments.append(SegmentPoly([(0, resheight), (resheight*1.4, 0), (0, -resheight)]))


class Schottky(Diode):
    ''' Schottky Diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        schottky_width = 0.1
        self.segments.append(Segment(
            [(resheight*1.4, resheight),
             (resheight*1.4-schottky_width, resheight),
             (resheight*1.4-schottky_width, resheight-schottky_width)]))
        self.segments.append(Segment(
            [(resheight*1.4, -resheight),
             (resheight*1.4+schottky_width, -resheight),
             (resheight*1.4+schottky_width, -resheight+schottky_width)]))


class DiodeTunnel(Diode):
    ''' Tunnel Diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        tunnel_width = 0.1
        self.segments.append(Segment([(resheight*1.4, resheight),
                                      (resheight*1.4-tunnel_width, resheight)]))
        self.segments.append(Segment([(resheight*1.4, -resheight),
                                      (resheight*1.4-tunnel_width, -resheight)]))


class DiodeShockley(Element2Term):
    ''' Shockley Diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(0, 0), (resheight*1.4, 0), (resheight*1.4, resheight),
                                      (resheight*1.4, -resheight), gap, (resheight*1.4, 0)]))
        self.segments.append(Segment([(0, -resheight), (0, resheight), (resheight*1.4, 0)]))


class Zener(Diode):
    ''' Zener Diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        zener_width = 0.1
        self.segments.append(Segment([(resheight*1.4, resheight),
                                      (resheight*1.4+zener_width, resheight+zener_width)]))
        self.segments.append(Segment([(resheight*1.4, -resheight),
                                      (resheight*1.4-zener_width, -resheight-zener_width)]))


class Varactor(Element2Term):
    ''' Varactor Diode/Varicap/Variable Capacitance Diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        capgap = .13
        self.segments.append(Segment([(0, 0), gap, 
                                     (resheight*1.4, resheight), (resheight*1.4, -resheight),
                                     gap,
                                     (resheight*1.4+capgap, resheight), (resheight*1.4+capgap, -resheight),
                                     gap,
                                     (resheight*1.4+capgap, 0)]))
        self.segments.append(SegmentPoly([(0, resheight), (resheight*1.4, 0), (0, -resheight)]))


class LED(Diode):
    ''' Light emitting diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(SegmentArrow((resheight, resheight*1.5),
                                          (resheight*2, resheight*3.25),
                                          headwidth=.16, headlength=.2))
        self.segments.append(SegmentArrow((resheight*.1, resheight*1.5),
                                          (resheight*1.1, resheight*3.25),
                                          headwidth=.16, headlength=.2))
        self.params['lblloc'] = 'bot'


class LED2(Diode):
    ''' Light emitting diode (curvy light lines) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        x = linspace(-1, 1)
        y = [-x0*(x0-.7)*(x0+.7)/2 + resheight*2.5 for x0 in x]
        x = linspace(resheight*.75, resheight*1.25)
        
        theta = 20
        p = [Point((x0, y0)).rotate(theta) for x0, y0 in zip(x, y)]
        p2 = [Point((x0-.2, y0)).rotate(theta) for x0, y0 in zip(x, y)]
        
        pa = Point((x[0], y[0]+.1)).rotate(theta)
        pa2 = Point((x[0]-.2, y[0]+.1)).rotate(theta)
        
        self.segments.append(Segment(p))
        self.segments.append(Segment(p2))
        self.segments.append(SegmentArrow(p[1], pa, headwidth=.15, headlength=.2))
        self.segments.append(SegmentArrow(p2[1], pa2, headwidth=.15, headlength=.2))
        self.params['lblloc'] = 'bot'


class Photodiode(Diode):
    ''' Photo-sensitive diode '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        x = linspace(-1, 1)
        y = [-x0*(x0-.7)*(x0+.7)/2 + resheight*2.5 for x0 in x]
        x = linspace(resheight*.75, resheight*1.25)
        theta = 20
        p = [Point((x0, y0)).rotate(theta) for x0, y0 in zip(x, y)]
        p2 = [Point((x0-.2, y0)).rotate(theta) for x0, y0 in zip(x, y)]

        pa = Point((x[-1], y[-1]-.1)).rotate(theta)
        pa2 = Point((x[-1]-.2, y[-1]-.1)).rotate(theta)
        self.segments.append(Segment(p))
        self.segments.append(Segment(p2))
        self.segments.append(SegmentArrow(p[-2], pa, headwidth=.15, headlength=.2))
        self.segments.append(SegmentArrow(p2[-2], pa2, headwidth=.15, headlength=.2))
        self.params['lblloc'] = 'bot'


class PotentiometerIEEE(ResistorIEEE):
    ''' Potentiometer (U.S. style)

        Anchors:
            tap
    '''
    # Ok, this has three terminals, but is works like a two-term with lead extension
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        potheight = .72
        self.anchors['tap'] = (reswidth*3, potheight)
        self.params['lblloc'] = 'bot'
        self.segments.append(SegmentArrow((reswidth*3, potheight), (reswidth*3, reswidth*1.5),
                                          headwidth=.15, headlength=.25))


class PotentiometerIEC(ResistorIEC):
    ''' Potentiometer (European style)

        Anchors:
            tap
    '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        potheight = .72
        self.anchors['tap'] = (reswidth*3, potheight)
        self.params['lblloc'] = 'bot'
        self.segments.append(SegmentArrow((reswidth*3, potheight), (reswidth*3, reswidth*2),
                                          headwidth=.15, headlength=.22))


class Diac(Element2Term):
    ''' Diac (diode for alternating current) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), gap, (resheight*1.4, resheight*1.8),
             (resheight*1.4, -resheight*1.8), gap,
             (0, resheight*1.8), (0, -resheight*1.8), gap, (resheight*1.4, 0)]))
        self.segments.append(SegmentPoly([(0, -resheight-.25), (resheight*1.4, -.25),
                                          (0, -resheight+.25)]))
        self.segments.append(SegmentPoly([(resheight*1.4, resheight+.25), (0, .25),
                                          (resheight*1.4, resheight-.25)]))


class Triac(Diac):
    ''' Triac

        Anchors:
            gate
    '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(resheight*1.4, .25), (resheight*1.4+.5, .5)]))
        self.anchors['gate'] = (resheight*1.4+.5, .5)


class SCR(Diode):
    ''' Silicon controlled rectifier (or thyristor)

        Anchors:
            gate
    '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(resheight*1.4, 0), (resheight*1.4+.3, -.3), (resheight*1.4+.3, -.5)]))
        self.anchors['gate'] = (resheight*1.4+.3, -.5)


class Memristor(Element2Term):
    ''' Memristor (resistor with memory) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        mr = 0.2
        self.segments.append(Segment(
            [(0, 0), (mr, 0), (mr, -mr*.75), (mr*2, -mr*.75), (mr*2, mr*.75),
             (mr*3, mr*.75), (mr*3, -mr*.75), (mr*4, -mr*.75), (mr*4, 0),
             (mr*5, 0)]))
        self.segments.append(Segment(
            [(0, mr*1.25), (mr*5, mr*1.25), (mr*5, mr*-1.25), (0, mr*-1.25),
             (0, mr*1.25)]))
        self.segments.append(SegmentPoly(
            [(0, mr*1.25), (0, -mr*1.25), (mr/2, -mr*1.25), (mr/2, mr*1.25)],
            fill='black'))


class Memristor2(Element2Term):
    ''' Memristor (resistor with memory), alternate style '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        mr = 0.2
        mrv = .25
        self.segments.append(Segment(
            [(0, 0), (0, mrv), (mr, mrv), (mr, -mrv), (mr*2, -mrv), (mr*2, mrv),
             (mr*3, mrv), (mr*3, -mrv), (mr*4, -mrv), (mr*4, mrv),
             (mr*5, mrv), (mr*5, -mrv), (mr*6, -mrv), (mr*6, 0),
             (mr*7, 0)]))


class Josephson(Element2Term):
    ''' Josephson Junction '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), gap, (-resheight, resheight), (resheight, -resheight),
             gap, (resheight, resheight), (-resheight, -resheight),
             gap, (0, 0)]))


class FuseUS(Element2Term):
    ''' Fuse (U.S. Style)

        Args:
            dots: Show dots on connections to fuse
    '''
    def __init__(self, *d, dots: bool=True, **kwargs):
        super().__init__(*d, **kwargs)
        fuser = .12
        fusex = linspace(fuser*2, 1+fuser)
        fusey = [math.sin(x) * resheight for x in linspace(0, 2*math.pi)]
        self.segments.append(Segment(list(zip(fusex, fusey))))
        self.segments.append(Segment([(0, 0), gap, (1+fuser*3, 0)]))
        if dots:
            self.fill(kwargs.get('fill', 'bg'))

    def fill(self, color: bool | str=True) -> 'Element':
        ''' Set element fill '''
        fuser = .12
        self.segments.append(SegmentCircle(
            (fuser, 0), fuser, zorder=4, fill=color))
        self.segments.append(SegmentCircle(
            (fuser*2+1, 0), fuser, zorder=4, fill=color))
        super().fill(color)
        return self


class FuseIEEE(ResistorIEC):
    ''' Fuse (IEEE Style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(0, 0), (reswidth*6, 0)]))


class FuseIEC(ResistorIEC):
    ''' Fuse (IEC Style) '''    
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        dx = resheight*.66
        self.segments.append(Segment([(dx, resheight), (dx, -resheight)]))
        self.segments.append(Segment([(6*reswidth-dx, resheight),
                                      (6*reswidth-dx, -resheight)]))            


class Breaker(Element2Term):
    ''' Circuit breaker

        Args:
            dots: Show connection dots
    '''
    def __init__(self, *d, dots: bool=True, **kwargs):
        super().__init__(*d, **kwargs)
        theta1 = 25 if dots else 10
        theta2 = 155 if dots else 170
        self.segments.append(Segment(
            [(0, 0), gap, (1, 0)]))
        self.segments.append(SegmentArc(
            (.5, 0), 1, .65, theta1=theta1, theta2=theta2))
        if dots:
            rad = .12
            self.segments.append(SegmentCircle((rad, 0), rad, zorder=4))
            self.segments.append(SegmentCircle((1-rad, 0), rad, zorder=4))


def cycloid(loops: int=4, ofst: Sequence[float]=(0, 0),
            a:float=.06, b:float=.19, norm:bool=True,
            vertical:bool=False,
            flip:bool=False) -> list[tuple[float, float]]:
    ''' Generate a prolate cycloid (inductor spiral) that
        will always start and end at y=0.

        Args:
            loops: Number of loops
            a, b: Parameters. b>a for prolate (loopy) cycloid
            norm: Normalize the length to 1
            vertical: draw vertically (swap x, y)
            flip: flip orientation (invert y)

        Returns:
            List of (x, y) coordinates defining the cycloid
    '''
    yint = math.acos(a/b)  # y-intercept
    t = linspace(yint, 2*(loops+1)*math.pi-yint, num=loops*50)
    x = [a*t0 - b*math.sin(t0) for t0 in t]
    y = [a - b*math.cos(t0) for t0 in t]
    x = [xx - x[0] for xx in x]  # Shift to start at 0,0

    if norm:
        x = [xx / (x[-1]-x[0]) for xx in x]  # Normalize length to 1

    if flip:
        y = [-yy for yy in y]

    y = [yy * (max(y)-min(y))/(resheight) for yy in y]  # Normalize to resistor width

    if vertical:
        x, y = y, x

    x = [xx + ofst[0] for xx in x]
    y = [yy + ofst[1] for yy in y]

    path = list(zip(x, y))
    return path


class Inductor(Element2Term):
    ''' Inductor '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        ind_w = .25
        self.segments.append(Segment([(0, 0), gap, (1, 0)]))
        for i in range(4):
            self.segments.append(SegmentArc(
                [(i*2+1)*ind_w/2, 0], theta1=0, theta2=180,
                width=ind_w, height=ind_w))


class Inductor2(Element2Term):
    ''' Inductor, drawn as cycloid (loopy)

        Args:
            loops: Number of inductor loops
    '''
    def __init__(self, *d, loops: int=4, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(cycloid(loops=loops)))


class CPE(Element2Term):
    ''' Constant Phase Element '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        capgap = 0.25
        offset = 0.25
        self.segments.append(Segment([(0, 0), (-offset, 0), gap, (offset, 0),
                                      (0, -resheight), gap, (capgap, 0)]))
        self.segments.append(Segment([(0, 0), (-offset, 0), gap, (offset, 0),
                                      (0, resheight), gap, (capgap, 0)]))
        self.segments.append(Segment([(0, 0), (-offset, resheight), gap,
                                      (0, -resheight), gap, (capgap, 0)]))
        self.segments.append(Segment([(0, 0), (-offset, -resheight), gap,
                                      (0, -resheight), gap, (capgap, 0)]))

        
        
# default to IEEE style
Resistor = ResistorIEEE
ResistorVar = ResistorVarIEEE
Photoresistor = PhotoresistorIEEE
Potentiometer = PotentiometerIEEE
Fuse = FuseIEEE

# Old names
RBox = ResistorIEC
RBoxVar = ResistorVarIEC
PotBox = PotentiometerIEC
PhotoresistorBox = PhotoresistorIEC


''' DEPRECATED element definitions, based on dictionaries instead of classes, translated
    into their newer class counterparts.
'''

from functools import partial

from .. import dsp
from ..elements.legacy import LINE, DOT, LINEDOT, LINE_DOT, ARROW, ARROWHEAD, ANT, ic, multiplexer
from ..flow import Box

def makebox(w, h):
    return partial(Box, w=w, h=h)

BOX = dsp.Square
CIRCLE = dsp.Circle
SUM = dsp.Sum
SUMSIGMA = dsp.SumSigma
MIX = dsp.Mixer
SPEAKER1 = dsp.Speaker
AMP = dsp.Amp
OSCBOX = dsp.OscillatorBox
OSC = dsp.Oscillator
FILT = dsp.Filter
FILT_BP = partial(dsp.Filter, response='bp')
FILT_LP = partial(dsp.Filter, response='lp')
FILT_HP = partial(dsp.Filter, response='hp')
ADC = dsp.Adc
DAC = dsp.Dac
DEMOD = dsp.Demod



''' One terminal element definitions '''

from ..segments import Segment
from .elements import Element
from .twoterm import resheight, gap

gndgap = 0.12
gnd_lead = 0.4


class Ground(Element):
    ''' Ground connection '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), (0, -gnd_lead), (-resheight, -gnd_lead),
             (resheight, -gnd_lead), gap, (-resheight*.7, -gndgap-gnd_lead),
             (resheight*.7, -gndgap-gnd_lead), gap,
             (-resheight*.2, -gndgap*2-gnd_lead),
             (resheight*.2, -gndgap*2-gnd_lead)]))
        self.params['theta'] = 0
        self.params['drop'] = (0, 0)
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)


class GroundSignal(Element):
    ''' Signal ground '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), (0, -gnd_lead), (-resheight, -gnd_lead), (0, -gnd_lead*2),
             (resheight, -gnd_lead), (0, -gnd_lead)]))
        self.params['drop'] = (0, 0)
        self.params['theta'] = 0
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)


class GroundChassis(Element):
    ''' Chassis ground '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        dx = resheight*.75
        dy = resheight
        self.segments.append(Segment(
            [(0, 0), (0, -gnd_lead), (-dx, -gnd_lead-dy)]))
        self.segments.append(Segment(
            [(0, -gnd_lead), (-dx, -gnd_lead), (-dx*2, -gnd_lead-dy)]))
        self.segments.append(Segment(
            [(0, -gnd_lead), (dx, -gnd_lead), (0, -gnd_lead-dy)]))
        self.params['drop'] = (0, 0)
        self.params['theta'] = 0
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)


class Antenna(Element):
    ''' Antenna '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        lead = 0.6
        h = 0.6
        w = 0.38
        self.segments.append(Segment(
            [(0, 0), (0, lead), (-w, lead+h), (w, lead+h), (0, lead)]))
        self.params['drop'] = (0, 0)
        self.params['theta'] = 0
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)


class AntennaLoop(Element):
    ''' Loop antenna (diamond style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        lead = 0.2
        h = 0.5
        self.segments.append(Segment(
            [(0, 0), (0, lead), (-h, h+lead), (lead/2, h*2+1.5*lead),
             (h+lead, h+lead), (lead, lead), (lead, 0)]))
        self.params['drop'] = (0, 0)
        self.params['theta'] = 0
        self.anchors['start'] = (0, 0)
        self.anchors['end'] = (lead, 0)


class AntennaLoop2(Element):
    ''' Loop antenna (square style) '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        lead = .25
        h = 1
        x1 = -h/2-lead/2
        x2 = x1 + lead
        x3 = h/2+lead/2
        x4 = x3 + lead
        y1 = lead
        y2 = lead*2
        y3 = h+y2
        y4 = y3+lead
        self.segments.append(Segment(
            [(0, 0), (0, y1), (x1, y1), (x1, y3), (x3, y3), (x3, y2),
             (x2, y2), (x2, y4), (x4, y4), (x4, y1), (lead, y1), (lead, 0)]))
        self.params['drop'] = (lead, 0)
        self.params['theta'] = 0
        self.anchors['start'] = (0, 0)
        self.anchors['end'] = (lead, 0)


class Vss(Element):
    ''' Vss connection '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        dx = resheight*.75
        self.segments.append(Segment([(0, 0), (0, -gnd_lead)]))
        self.segments.append(Segment([(0, -gnd_lead), (-dx, -gnd_lead)]))
        self.segments.append(Segment([(0, -gnd_lead), (dx, -gnd_lead)]))
        self.params['drop'] = (0, 0)
        self.params['theta'] = 0
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)


class Vdd(Element):
    ''' Vdd connection '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        dx = resheight*.75
        self.segments.append(Segment([(0, 0), (0, gnd_lead)]))
        self.segments.append(Segment([(0, gnd_lead), (-dx, gnd_lead)]))
        self.segments.append(Segment([(0, gnd_lead), (dx, gnd_lead)]))
        self.params['drop'] = (0, 0)
        self.params['theta'] = 0
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)



''' Data types for schemdraw '''

import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from typing import Union, Tuple, Sequence, Optional
from collections import namedtuple
from enum import Enum, unique

from .util import Point

BBox = namedtuple('BBox', ['xmin', 'ymin', 'xmax', 'ymax'])

# matplotlib uses 'projecting' to be the same as 'square' in svg.
Capstyle = Literal['butt', 'round', 'square', 'projecting']
Joinstyle = Literal['bevel', 'miter', 'round']
Linestyle = Literal['-', ':', '--', '-.']
Direction = Union[Literal['up', 'down', 'left', 'right',
                          'u', 'd', 'l', 'r'], int]
Halign = Literal['center', 'left', 'right']
Valign = Literal['center', 'top', 'bottom']
Align = Tuple[Optional[Halign], Optional[Valign]]
Arcdirection = Literal['cw', 'ccw']
Side = Literal['top', 'bot', 'lft', 'rgt', 'bottom', 'left', 'right', 'L', 'R', 'T', 'B']
LabelLoc = Union[Side, str]
XY = Union[Sequence[float], Point]
RotationMode = Literal['anchor', 'default']
TextMode = Literal['path', 'text']

BilateralDirection = Literal['in', 'out']
EndRef = Literal['start', 'end']
ActionType = Optional[Literal['open', 'close']]
HeaderStyle = Literal['round', 'square', 'screw']
HeaderNumbering = Literal['lr', 'ud', 'ccw']


Backends = Literal['svg', 'matplotlib']


@unique
class ImageFormat(str, Enum):
    ''' Known Matplotlib image formats '''
    EPS = 'eps'
    JPG = 'jpg'
    PDF = 'pdf'
    PGF = 'pgf'
    PNG = 'png'
    PS = 'ps'
    RAW = 'raw'
    RGBA = 'rgba'
    SVG = 'svg'
    TIF = 'tif'


ImageType = Literal['eps', 'jpg', 'pdf', 'pgf', 'png', 'ps',
                    'raw', 'rgba', 'svg', 'tif']



''' Schemdraw Drawing class '''

from __future__ import annotations
from typing import Type, Any
from collections import ChainMap
import warnings
import math

from .types import BBox, Backends, ImageFormat, Linestyle, Arcdirection, XY, ImageType, BilateralDirection
from .elements import Element, _set_elm_backend
from .elements.lines import LoopCurrent, CurrentLabel, CurrentLabelInline
from .segments import Segment, SegmentText, SegmentArc, SegmentArrow, SegmentCircle, SegmentPoly
from .util import Point

from .backends.svg import Figure as svgFigure
try:
    from .backends.mpl import Figure as mplFigure
except ImportError:
    mplFigure = None  # type: ignore


Figure: Type[svgFigure] | Type[mplFigure]
if mplFigure is None:
    Figure = svgFigure
else:
    Figure = mplFigure
_set_elm_backend(Figure)


def use(backend: Backends='matplotlib') -> None:
    ''' Change default backend, either 'matplotlib' or 'svg' '''
    global Figure
    if backend == 'matplotlib':
        if mplFigure is None:
            raise ValueError('Could not import Matplotlib.')
        Figure = mplFigure
    else:
        Figure = svgFigure
    _set_elm_backend(Figure)


def config(unit: float=3.0, inches_per_unit: float=0.5,
           lblofst: float=0.1, fontsize: float=14,
           font: str='sans-serif', color: str='black',
           lw: float=2, ls: Linestyle='-',
           fill: str=None, bgcolor: str=None) -> None:
    ''' Set global schemdraw style configuration
 
        Args:
            unit: Full length of a 2-terminal element. Inner zig-zag portion
                of a resistor is 1.0 units.
            inches_per_unit: Inches per drawing unit for setting drawing scale
            lblofst: Default offset between element and its label
            fontsize: Default font size for text labels
            font: Default font family for text labels
            color: Default color name or RGB (0-1) tuple
            lw: Default line width for elements
            ls: Default line style
            fill: Deault fill color for closed elements
    '''
    schemdrawstyle['unit'] = unit
    schemdrawstyle['inches_per_unit'] = inches_per_unit
    schemdrawstyle['lblofst'] = lblofst
    schemdrawstyle['fontsize'] = fontsize
    schemdrawstyle['font'] = font
    schemdrawstyle['color'] = color
    schemdrawstyle['lw'] = lw
    schemdrawstyle['ls'] = ls
    schemdrawstyle['fill'] = fill 
    if bgcolor:
        schemdrawstyle['bgcolor'] = bgcolor


schemdrawstyle: dict[str, Any] = {}  # Global style
config()  # Initialize default configuration


def theme(theme='default'):
    ''' Set schemdraw theme (line color and background color).
        Themes match those in jupyter-themes package
        (https://github.com/dunovank/jupyter-themes).

        Available themes:
            * default (black on white)
            * dark (white on black)
            * solarizedd
            * solarizedl
            * onedork
            * oceans16
            * monokai
            * gruvboxl
            * gruvboxd
            * grade3
            * chesterish
    '''
    if theme == 'default':
        config(bgcolor='white')
    elif theme == 'dark':
        schemdrawstyle['color'] = 'white'
        schemdrawstyle['bgcolor'] = 'black'
    elif theme == 'solarizedd':
        schemdrawstyle['bgcolor'] = '#002b36'
        schemdrawstyle['color'] = '#657b83'
    elif theme == 'solarizedl':
        schemdrawstyle['bgcolor'] = '#eee8d5'
        schemdrawstyle['color'] = '#073642'
    elif theme == 'onedork':
        schemdrawstyle['bgcolor'] = '#373e4b'
        schemdrawstyle['color'] = '#899ab8'
    elif theme == 'oceans16':
        schemdrawstyle['bgcolor'] = '#384151'
        schemdrawstyle['color'] = '#CDD2E9'
    elif theme == 'monokai':
        schemdrawstyle['bgcolor'] = '#232323'
        schemdrawstyle['color'] = '#BBBBBB'
    elif theme == 'gruvboxl':
        schemdrawstyle['bgcolor'] = '#ebdbb2'
        schemdrawstyle['color'] = '#3c3836'
    elif theme == 'gruvboxd':
        schemdrawstyle['bgcolor'] = '#1d2021'
        schemdrawstyle['color'] = '#d5c4a1'
    elif theme == 'grade3':
        schemdrawstyle['bgcolor'] = '#ffffff'
        schemdrawstyle['color'] = '#3f3d46'
    elif theme == 'chesterish':
        schemdrawstyle['bgcolor'] = '#323A48'
        schemdrawstyle['color'] = '#92A2BD'
    else:
        raise ValueError(f'Unknown theme {theme}')


class Drawing:
    ''' A schematic drawing

        See `schemdraw.config` method for argument defaults

        Args:
            *elements: List of Element instances to add to the drawing
            unit: Full length of a 2-terminal element. Inner zig-zag portion
                of a resistor is 1.0 units.
            inches_per_unit: Inches per drawing unit for setting drawing scale
            lblofst: Default offset between element and its label
            fontsize: Default font size for text labels
            font: Default font family for text labels
            color: Default color name or RGB (0-1) tuple
            lw: Default line width for elements
            ls: Default line style
            fill: Deault fill color for closed elements

        Attributes:
            here: (xy tuple) Current drawing position. The next element will
                be added at this position unless specified otherwise.
            theta: (float) Current drawing angle, in degrees. The next
                element will be added with this angle unless specified
                otherwise.
    '''
    def __init__(self, *elements: Element, unit: float=None,
                 inches_per_unit: float=None, lblofst: float=None,
                 fontsize: float=None, font: str=None,
                 color: str=None, lw: float=None, ls: Linestyle=None,
                 fill: str=None):
        self.elements: list[Element] = []
        self.inches_per_unit = inches_per_unit if inches_per_unit is not None else schemdrawstyle.get('inches_per_unit')
        self.unit = unit if unit is not None else schemdrawstyle.get('unit')
        
        self.dwgparams: dict[str, Any] = {}
        if unit:
            self.dwgparams['unit'] = unit
        if font:
            self.dwgparams['font'] = font
        if fontsize:
            self.dwgparams['fontsize'] = fontsize
        if lblofst:
            self.dwgparams['lblofst'] = lblofst
        if color:
            self.dwgparams['color'] = color
        if ls:
            self.dwgparams['ls'] = ls
        if lw:
            self.dwgparams['lw'] = lw
        if fill:
            self.dwgparams['fill'] = fill

        self.here: XY = Point((0, 0))
        self.theta: float = 0
        self._state: list[tuple[Point, float]] = []  # Push/Pop stack
        self._interactive = False
        self.fig = None

        for element in elements:
            self.add(element)

    def interactive(self, interactive: bool=True):
        ''' Enable interactive mode (matplotlib backend only). Matplotlib
            must also be set to interactive with `plt.ion()`.
        '''
        self._interactive = interactive

    def get_bbox(self) -> BBox:
        ''' Get drawing bounding box '''
        xmin = math.inf
        xmax = -math.inf
        ymin = math.inf
        ymax = -math.inf
        for element in self.elements:
            bbox = element.get_bbox(transform=True)
            xmin = min(bbox.xmin, xmin)
            xmax = max(bbox.xmax, xmax)
            ymin = min(bbox.ymin, ymin)
            ymax = max(bbox.ymax, ymax)
        return BBox(xmin, ymin, xmax, ymax)

    def get_segments(self) -> list['Segment' | 'SegmentText' | 'SegmentArc' | 'SegmentArrow' |
                                   'SegmentCircle' | 'SegmentPoly']:
        ''' Get flattened list of all segments in the drawing '''
        segments = []
        for element in self.elements:
            segments.extend([s.xform(element.transform, **element._cparams)
                             for s in element.segments])
        return segments

    def _repr_svg_(self):
        ''' SVG representation for Jupyter '''
        return self.draw().getimage('svg').decode()

    def _repr_png_(self):
        ''' PNG representation for Jupyter '''
        if Figure == mplFigure:
            return self.draw().getimage('png')
        return None

    def __iadd__(self, element: Element):
        ''' In-place add element, via += operator. '''
        self.add(element)
        return self

    def add(self, element: Element | Type[Element], **kwargs) -> Element:
        ''' Add an element to the drawing.

            Args:
                element: The element to add.
        '''
        # TODO: remove kwargs when deprecated dictionary elements are removed
        if not isinstance(element, Element):
            # Instantiate it (for support of legacy add method)
            element = element(**kwargs)
        elif len(kwargs) > 0:
            warnings.warn('kwargs to add method are ignored because element is already instantiated')

        dwgparams = ChainMap(self.dwgparams, schemdrawstyle)
        self.here, self.theta = element._place(self.here, self.theta,
                                               **dwgparams)
        self.elements.append(element)
        
        if self._interactive:
            if self.fig is None:
                self._initfig()
            element._draw(self.fig)
            self.fig.set_bbox(self.get_bbox())  # type: ignore
            self.fig.getimage()  # type: ignore
        else:
            self.fig = None  # Clear any existing figure
        return element

    def add_elements(self, *elements: Element) -> None:
        ''' Add multiple elements to the drawing '''
        for element in elements:
            self.add(element)

    def undo(self) -> None:
        ''' Removes previously added element '''
        self.elements.pop(-1)
        self.fig.clear()  # type: ignore
        for element in self.elements:
            element._draw(self.fig)
        self.here, self.theta = self.elements[-1].absdrop
        self.fig.set_bbox(self.get_bbox())  # type: ignore
        self.fig.getimage()  # type: ignore

    def move(self, dx: float, dy: float) -> None:
        ''' Move the current drawing position

            Args:
                dx: change in x position
                dy: change in y position
        '''
        self.here = Point((self.here[0] + dx, self.here[1] + dy))

    def push(self) -> None:
        ''' Push/save the drawing state.
            Drawing.here and Drawing.theta are saved.
        '''
        self._state.append((Point(self.here), self.theta))

    def pop(self) -> None:
        ''' Pop/load the drawing state. Location and angle are returned to
            previously pushed state.
        '''
        if len(self._state) > 0:
            self.here, self.theta = self._state.pop()

    def loopI(self, elm_list: list[Element], label: str='',
              d: Arcdirection='cw', theta1: float=35,
              theta2: float=-35, pad: float=.2, color: str=None) -> Element:
        ''' Draw an arc to indicate a loop current bordered by elements in list

            [DEPRECATED - use LoopCurrent element instead]

            Args:
                elm_list: Boundary elements in order of top, right, bot, left
                label: Text label to draw in center of loop
                d: Arc/arrow direction
                theta1: Start angle of arrow arc (degrees)
                theta2: End angle of arrow arc (degrees)
                pad: Distance between elements and arc
                color: Color for loop arrow
        '''
        warnings.warn('loopI function is deprecated. Add a LoopCurrent element instead.', DeprecationWarning)

        element = LoopCurrent(elm_list, theta1=theta1, theta2=theta2,
                              direction=d, pad=pad)
        element.label(label)
        if color:
            element.color(color)
        self.add(element)
        return element

    def labelI(self, elm: Element, label: str='', arrowofst: float=0.4,
               arrowlen: float=2, reverse: bool=False, top: bool=True,
               color: str=None) -> Element:
        ''' Add an arrow element along side another element

            [DEPRECATED - use CurrentLabel element instead]

            Args:
                elm: Element to add arrow to
                label: String or list of strings to evenly space along arrow
                arrowofst: Distance from element to arrow
                arrowlen: Length of arrow as multiple of Drawing.unit
                reverse: Reverse the arrow direction
                top: Draw arrow on top (True) or bottom (False) of element
                color: Color for label. Defaults to color of elm.
        '''
        warnings.warn('labelI function is deprecated. Add a CurrentLabel element instead.', DeprecationWarning)
        element = CurrentLabel(ofst=arrowofst, length=arrowlen,
                               top=top, reverse=reverse)
        element.at(elm.center).theta(elm.transform.theta)
        element.label(label)
        if color:
            element.color(color)
        self.add(element)
        return element

    def labelI_inline(self, elm: Element, label: str=None,
                      botlabel: str=None, d: BilateralDirection='in',
                      start: bool=True, ofst: float=.8,
                      color: str=None) -> Element:
        ''' Add an arrowhead for labeling current inline with leads.
            Works on Element2Term elements.

            [DEPRECATED - use CurrentLabelInline element instead]

            Args:
                elm: Element to add arrow to
                label: Text to draw above the arrowhead
                botlabel: Text to draw below the arrowhead
                d: Arrowhead direction, into or out of the element
                start: Place arrowhead near start (True) or
                    end (False) of element
                ofst: Offset from center of element
                color: Color for label. Defaults to color of elm.
        '''
        warnings.warn('labelI_inline function is deprecated. Add a CurrentLabelInline element instead.', DeprecationWarning)
        element = CurrentLabelInline(direction=d, ofst=ofst, start=start)
        element.at(elm.center).theta(elm.transform.theta).zorder(3)
        if label:
            element.label(label)
        if botlabel:
            element.label(botlabel, 'bot')
        if color:
            element.color(color)
        self.add(element)
        return element

    def _initfig(self, ax=None, backend: Backends=None, showframe: bool=False) -> None:
        figclass = {'matplotlib': mplFigure,
                    'svg': svgFigure}.get(backend, Figure)  # type: ignore
        fig = figclass(ax=ax,
                       bbox=self.get_bbox(),
                       inches_per_unit=self.inches_per_unit,
                       showframe=showframe)
        
        params = ChainMap(self.dwgparams, schemdrawstyle)
        if 'bgcolor' in params:
            fig.bgcolor(params['bgcolor'])
        self.fig = fig
    
    def draw(self, showframe: bool=False, show: bool=True,
             ax=None, backend: Backends=None):
        ''' Draw the schematic

            Args:
                showframe: Show axis frame. Useful for debugging a drawing.
                show: Show the schematic in a GUI popup window (when
                    outside of a Jupyter inline environment)
                ax: Existing axis to draw on. Should be set to equal
                    aspect for best results.

            Returns:
                schemdraw Figure object
        '''
        # Redraw only if needed. Otherwise, show/return what's already
        # been drawn on the figure
        if (self.fig is None or
            ax is not None or showframe != self.fig.showframe or
            backend is not None):
            self._initfig(ax=ax, backend=backend, showframe=showframe)
            for element in self.elements:
                element._draw(self.fig)

        if show:
            # Show figure in window if not inline/Jupyter mode            
            self.fig.show()  # type: ignore
        return self.fig  # Otherwise return Figure and let _repr_ display it

    def save(self, fname: str, transparent: bool=True, dpi: float=72) -> None:
        ''' Save figure to a file

            Args:
                fname: Filename to save. File type automatically determined
                    from extension (png, svg, jpg)
                transparent: Save as transparent background, if available
                dpi: Dots-per-inch for raster formats
        '''
        if self.fig is None:
            self.draw(show=False)
        self.fig.save(fname, transparent=transparent, dpi=dpi)  # type: ignore

    def get_imagedata(self, fmt: ImageFormat | ImageType) -> bytes:
        ''' Get image data as bytes array

            Args:
                fmt: Format or file extension of the image type

            Returns:
                Image data as bytes
        '''
        if Figure == svgFigure and fmt.lower() != 'svg':
            raise ValueError('Format not available in SVG backend.')
        if self.fig is None:
            self.draw(show=False)
        return self.fig.getimage(ext=fmt)  # type: ignore



''' Matplotlib drawing backend for schemdraw '''

from typing import Sequence
from io import BytesIO
import math

import matplotlib  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
from matplotlib.patches import Arc  # type: ignore

from .. import util
from ..types import Capstyle, Joinstyle, Linestyle, BBox

inline = 'inline' in matplotlib.get_backend()


def fix_capstyle(capstyle):
    ''' Matplotlib uses 'projecting' rather than 'square' for some reason '''
    if capstyle == 'square':
        return 'projecting'
    return capstyle


class Figure:
    ''' Schemdraw figure on Matplotlib figure

        Args:
            bbox: Coordinate bounding box for drawing, used to
                override Matplotlib's autoscale
            inches_per_unit: Scale for the drawing
            showframe: Show Matplotlib axis frame
            ax: Existing Matplotlib axis to draw on
    '''
    def __init__(self, **kwargs):
        self.showframe = kwargs.get('showframe', False)
        self.bbox = kwargs.get('bbox', None)
        self.inches_per_unit = kwargs.get('inches_per_unit', .5)        
        if kwargs.get('ax'):
            self.ax = kwargs.get('ax')
            self.fig = self.ax.figure
            self.userfig = True
        else:
            # a Figure (big F) is not part of the pyplot interface
            # so won't be shown double in Jupyter/inline interfaces.
            # But a figure (small f) is required to show the image in
            # MPL's popup GUI.
            if inline:
                self.fig = plt.Figure()
            else:
                self.fig = plt.figure()
            self.fig.subplots_adjust(
                left=0.05,
                bottom=0.05,
                right=0.95,
                top=0.90)
            self.ax = self.fig.add_subplot()
            self.userfig = False
            self.ax.set_aspect('equal')
            if not self.showframe:
                self.ax.axes.get_xaxis().set_visible(False)
                self.ax.axes.get_yaxis().set_visible(False)
                self.ax.set_frame_on(False)
        self.ax.autoscale_view(True)  # This autoscales all the shapes too

    def set_bbox(self, bbox: BBox):
        ''' Set bounding box, to override Matplotlib's autoscale '''
        self.bbox = bbox

    def show(self) -> None:
        ''' Display figure in interactive window. Does nothing
            when running inline (ie Jupyter) which shows the
            figure using _repr_ methods.
        '''
        if not inline:
            self.getfig()
            self.fig.show()
            plt.show()   # To start the MPL event loop

    def bgcolor(self, color: str) -> None:
        ''' Set background color of drawing '''
        self.fig.set_facecolor(color)

    def plot(self, x: float, y: float, color: str='black', ls: Linestyle='-',
             lw: float=2, fill: str=None, capstyle: Capstyle='round',
             joinstyle: Joinstyle='round',  zorder: int=2) -> None:
        ''' Plot a path '''
        self.ax.plot(x, y, zorder=zorder, color=color, ls=ls, lw=lw,
                     solid_capstyle=fix_capstyle(capstyle),
                     solid_joinstyle=joinstyle)
        if fill:
            self.ax.fill(x, y, color=fill, zorder=zorder)

    def text(self, s: str, x, y, color='black', fontsize=14, fontfamily='sans-serif',
             rotation=0, halign='center', valign='center', rotation_mode='anchor',
             zorder=3) -> None:
        ''' Add text to the figure '''
        self.ax.text(x, y, s, transform=self.ax.transData, color=color,
                     fontsize=fontsize, fontfamily=fontfamily,
                     rotation=rotation, rotation_mode=rotation_mode,
                     horizontalalignment=halign, verticalalignment=valign,
                     zorder=zorder)

    def poly(self, verts: Sequence[Sequence[float]], closed: bool=True,
             color: str='black', fill: str=None, lw: float=2, ls: Linestyle='-',
             capstyle: Capstyle='round', joinstyle: Joinstyle='round', zorder: int=1) -> None:
        ''' Draw a polynomial '''
        p = plt.Polygon(verts, closed=closed, ec=color,
                        fc=fill, fill=fill is not None,
                        lw=lw, ls=ls, capstyle=fix_capstyle(capstyle),
                        joinstyle=joinstyle, zorder=zorder)
        self.ax.add_patch(p)

    def circle(self, center: Sequence[float], radius: float, color: str='black', fill: str=None,
               lw: float=2, ls: Linestyle='-', zorder: int=1) -> None:
        ''' Draw a circle '''
        circ = plt.Circle(xy=center, radius=radius, ec=color, fc=fill,
                          fill=fill is not None, lw=lw, ls=ls, zorder=zorder)
        self.ax.add_patch(circ)

    def arrow(self, x: float, y: float, dx: float, dy: float,
              headwidth: float=.2, headlength: float=.2,
              color: str='black', lw: float=2, zorder: int=1) -> None:
        ''' Draw an arrow '''
        # Easier to skip Matplotlib's arrow or annotate methods and just draw a line
        # and a polygon.
        head = util.Point((x+dx, y+dy))
        tail = util.Point((x, y))
        fullen = math.sqrt(dx**2 + dy**2)
        theta = math.degrees(math.atan2(dy, dx))

        # Endpoints of the arrow fins
        fin1 = util.Point((fullen - headlength, headwidth/2)).rotate(theta) + tail
        fin2 = util.Point((fullen - headlength, -headwidth/2)).rotate(theta) + tail

        # Shorten the line so it doesn't poke through the arrowhead
        linehead = util.Point((dx - headlength * math.cos(math.radians(theta)) + tail.x,
                              dy - headlength * math.sin(math.radians(theta)) + tail.y))

        self.ax.plot([tail.x, linehead.x], [tail.y, linehead.y],
                lw=lw, color=color)
        p = plt.Polygon((fin1, head, fin2), closed=True, ec='none',
                        fc=color, fill=color is not None,
                        lw=1, zorder=zorder)
        self.ax.add_patch(p)


    def arc(self, center: Sequence[float], width: float, height: float,
            theta1: float=0, theta2: float=90, angle: float=0,
            color: str='black', lw: float=2, ls: Linestyle='-',
            zorder: int=1, arrow: bool=None) -> None:
        ''' Draw an arc or ellipse, with optional arrowhead '''
        arc = Arc(center, width=width, height=height, theta1=theta1,
                  theta2=theta2, angle=angle, color=color,
                  lw=lw, ls=ls, zorder=zorder)
        self.ax.add_patch(arc)

        if arrow is not None:
            x, y = math.cos(math.radians(theta2)), math.sin(math.radians(theta2))
            th2 = math.degrees(math.atan2((width/height)*y, x))
            x, y = math.cos(math.radians(theta1)), math.sin(math.radians(theta1))
            th1 = math.degrees(math.atan2((width/height)*y, x))
            if arrow == 'ccw':
                dx = math.cos(math.radians(th2+90)) / 100
                dy = math.sin(math.radians(theta2+90)) / 100
                s = util.Point((center[0] + width/2*math.cos(math.radians(th2)),
                                center[1] + height/2*math.sin(math.radians(th2))))
            else:
                dx = -math.cos(math.radians(th1+90)) / 100
                dy = -math.sin(math.radians(th1+90)) / 100

                s = util.Point((center[0] + width/2*math.cos(math.radians(th1)),
                                center[1] + height/2*math.sin(math.radians(th1))))

            s = util.rotate(s, angle, center)
            darrow = util.rotate((dx, dy), angle)

            self.ax.arrow(s[0], s[1], darrow[0], darrow[1], head_width=.15,
                          head_length=.25, color=color, zorder=zorder)

    def save(self, fname: str, transparent: bool=True, dpi: float=72) -> None:
        ''' Save the figure to a file '''
        fig = self.getfig()
        fig.savefig(fname, bbox_inches='tight', transparent=transparent, dpi=dpi,
                    bbox_extra_artists=self.ax.get_default_bbox_extra_artists())

    def getfig(self):
        ''' Get the Matplotlib figure '''
        if not self.userfig:
            if self.bbox is None or math.inf in self.bbox or -math.inf in self.bbox:
                # Use MPL's bbox, which sometimes clips things like arrowheads
                x1, x2 = self.ax.get_xlim()
                y1, y2 = self.ax.get_ylim()
            else:
                x1, y1, x2, y2 = self.bbox
            x1 -= .1  # Add a bit to account for line widths getting cut off
            x2 += .1
            y1 -= .1
            y2 += .1
            try:
                self.ax.set_xlim(x1, x2)
                self.ax.set_ylim(y1, y2)
            except ValueError:
                pass  # No elements
            w = x2-x1
            h = y2-y1

            if not self.showframe:
                self.ax.axes.get_xaxis().set_visible(False)
                self.ax.axes.get_yaxis().set_visible(False)
                self.ax.set_frame_on(False)
            try:
                self.ax.get_figure().set_size_inches(self.inches_per_unit*w,
                                                     self.inches_per_unit*h)
            except ValueError:
                pass  # infinite size (no elements yet)
        return self.fig

    def getimage(self, ext='svg'):
        ''' Get the image as SVG or PNG bytes array '''
        fig = self.getfig()
        output = BytesIO()
        fig.savefig(output, format=ext, bbox_inches='tight')
        return output.getvalue()

    def clear(self) -> None:
        ''' Remove everything '''
        self.ax.clear()
    
    def __repr__(self):
        if plt.isinteractive():
            self.show()
        return super().__repr__()
    
    def _repr_png_(self):
        ''' PNG representation for Jupyter '''
        return self.getimage('png')

    def _repr_svg_(self):
        ''' SVG representation for Jupyter '''
        return self.getimage('svg').decode()



from .flow import Box, Subroutine, Data, Start, Decision, Connect
from ..elements import Arrow, Arrowhead, Line, Dot


from . import legacy
import warnings

def __getattr__(name):
    e = getattr(legacy, name, None)
    if e is None:
        raise AttributeError('Element `{}` not found.'.format(name))
    warnings.warn('Dictionary-based elements are deprecated. Update to class-based elements or import from schemdraw.flow.legacy.', DeprecationWarning)
    return e


''' Utility functions for point geometry '''

from __future__ import annotations
from typing import Sequence

import math
from operator import mul
from itertools import starmap


def dot(a: Sequence[float], b: Sequence[Sequence[float]]) -> 'Point':
    ''' Dot product of iterables a and b '''
    return Point([sum(starmap(mul, zip(a, col))) for col in zip(*b)])


def linspace(start: float, stop: float, num: int=50) -> list[float]:
    ''' List of evenly spaced numbers '''
    step = (stop - start) / (num - 1)
    return [start+step*i for i in range(num)]


def rotate(xy: Sequence[float],
           angle: float,
           center: Sequence[float]=(0, 0)) -> 'Point':
    ''' Rotate the xy point by angle degrees '''
    co = math.cos(math.radians(angle))
    so = math.sin(math.radians(angle))
    center = Point(center)
    m = [[co, so], [-so, co]]  # rotation matrix
    b = Point(xy) - center
    b = dot(b, m)
    b = b + center
    return b


def mirrorx(xy, centerx=0) -> tuple[float, float]:
    ''' Mirror the point horizontally '''
    return -(xy[0]-centerx)+centerx, xy[1]


def flip(xy: Sequence[float]) -> tuple[float, float]:
    ''' Flip the point vertically '''
    return xy[0], -xy[1]


def delta(a: Sequence[float], b: Sequence[float]) -> tuple[float, float]:
    ''' Delta between points a and b '''
    return b[0] - a[0], b[1] - a[1]


def angle(a: Sequence[float], b: Sequence[float]) -> float:
    ''' Compute angle from point a to b '''
    theta = math.degrees(math.atan2(b[1] - a[1], b[0] - a[0]))
    return theta


def dist(a: Sequence[float], b: Sequence[float]) -> float:
    ''' Get distance from point a to b.

        Same as math.dist in Python 3.8+.
    '''
    try:
        # Python 3.8+
        return math.dist(a, b)  # type: ignore
    except AttributeError:
        return math.sqrt((a[0]-b[0])**2 + (a[1]-b[1])**2)


class Point(tuple):
    ''' An (x, y) tuple that can do math operations '''
    @property
    def x(self) -> float:
        ''' X value of point '''
        return self[0]

    @property
    def y(self) -> float:
        ''' Y value of point '''
        return self[1]

    def __repr__(self):
        return f'Point({self.x},{self.y})'

    def __add__(self, a):
        try:
            return Point((self.x+a.x, self.y+a.y))
        except AttributeError:
            return Point((self.x+a, self.y+a))

    def __sub__(self, a):
        try:
            return Point((self.x-a.x, self.y-a.y))
        except AttributeError:
            return Point((self.x-a, self.y-a))

    def __rsub__(self, a):
        try:
            return Point((a.x-self.x, a.y-self.y))
        except AttributeError:
            return Point((a-self.x, a-self.y))

    def __mul__(self, a: float):
        return Point((a*self.x, a*self.y))

    def __truediv__(self, a: float):
        return Point((self.x/a, self.y/a))

    def __neg__(self):
        return Point((-self.x, -self.y))

    __radd__ = __add__
    __rmul__ = __mul__

    def rotate(self, angle: float, center: Sequence[float] = (0, 0)) -> 'Point':
        ''' Rotate the point by angle degrees about the center '''
        return Point(rotate(self, angle, center=Point(center)))

    def mirrorx(self, centerx: float=0) -> 'Point':
        ''' Mirror in x direction about the centerx point '''
        return Point(mirrorx(self, centerx))

    def flip(self) -> 'Point':
        ''' Flip the point vertically '''
        return Point(flip(self))



''' Cable elements, coaxial and triaxial '''

import math
import warnings

from ..segments import Segment, SegmentArc
from .elements import Element2Term, gap


class Coax(Element2Term):
    ''' Coaxial cable element.

        Args:
            length: Total length of the cable, excluding lead extensions.
            radius: Radius of shield
            leadlen: Distance (x) from start of center conductor to
                start of shield.

        Anchors:
            * shieldstart
            * shieldstart_top
            * shieldend
            * shieldend_top
            * shieldcenter
            * shieldcenter_top
    '''
    def __init__(self, *d, length: float=3, radius: float=0.3, leadlen: float=0.6, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(    # Center conductor
                [(0, 0), (leadlen, 0), gap, (length-leadlen+radius/2, 0),
                 (length, 0)]))
        self.segments.append(Segment(
            [(leadlen, radius), (length-leadlen, radius)]))   # Top
        self.segments.append(Segment(
            [(leadlen, -radius), (length-leadlen, -radius)]))  # Bottom
        self.segments.append(SegmentArc(
            (leadlen, 0), width=radius, height=radius*2, theta1=0, theta2=360))
        self.segments.append(SegmentArc(
            (length-leadlen, 0), width=radius,
            height=radius*2, theta1=270, theta2=90))
        self.anchors['shieldstart'] = (leadlen, -radius)
        self.anchors['shieldstart_top'] = (leadlen, radius)
        self.anchors['shieldend'] = (length-leadlen, -radius)
        self.anchors['shieldend_top'] = (length-leadlen, radius)
        self.anchors['shieldcenter'] = (length/2, -radius)
        self.anchors['shieldcenter_top'] = (length/2, radius)

        if radius/2 > leadlen:
            warnings.warn('Coax leadlen < radius/2. Coax may be malformed.')
        if leadlen*2 > length:
            warnings.warn('Coax length < 2*leadlen. Coax may be malformed.')


class Triax(Element2Term):
    ''' Triaxial cable element.

        Args:
            length: Total length of the cable
            radiusinner: Radius of inner guard
            radiusouter: Radius of outer shield
            leadlen: Distance (x) from start of center conductor to
                start of guard.
            shieldofststart: Distance from start of inner guard to start
                of outer shield
            shieldofstend: Distance from end of outer shield to end
                of inner guard

        Anchors:
            * shieldstart
            * shieldstart_top
            * shieldend
            * shieldend_top
            * shieldcenter
            * shieldcenter_top
            * guardstart
            * guardstart_top
            * guardend
            * guardend_top
    '''
    def __init__(self, *d,
                 length: float=3,
                 leadlen: float=0.6,
                 radiusinner: float=0.3,
                 radiusouter: float=0.6,
                 shieldofststart: float=0.3,
                 shieldofstend: float=0.3,
                 **kwargs):
        super().__init__(*d, **kwargs)
        if radiusouter < radiusinner:
            raise ValueError('Triax inner radius > outer radius')

        xshield = radiusouter/2 * math.sqrt(1 - radiusinner**2/radiusouter**2)
        if shieldofststart - xshield > -radiusinner/2:
            thetashield = 180 - math.degrees(math.atan2(radiusinner, xshield))
        else:
            thetashield = 180

        if length - leadlen - shieldofstend + radiusouter < length:
            # Include the inner guard on output side
            self.segments.append(Segment(  # Center conductor (first)
                [(0, 0), (leadlen, 0), gap,
                 (length-leadlen+radiusinner/2, 0),
                 (length, 0)]))
            self.segments.append(Segment(
                [(length-leadlen-shieldofstend+xshield, radiusinner),
                 (length-leadlen, radiusinner)]))  # guard (inner) top/right
            self.segments.append(Segment(
                [(length-leadlen-shieldofstend+xshield, -radiusinner),
                 (length-leadlen, -radiusinner)]))  # guard (inner) bottom/right
            self.segments.append(SegmentArc(
                (length-leadlen, 0), width=radiusinner, height=radiusinner*2,
                theta1=270, theta2=90))
        else:
            # Don't include inner guard on output side
            self.segments.append(Segment(    # Center conductor
                [(0, 0), (leadlen, 0), gap,
                 (length-leadlen-shieldofstend+radiusouter/2, 0)]))

        # Start with shapes that are always shown...
        self.segments.append(Segment(  # guard (inner) top/left
            [(leadlen, radiusinner), (leadlen+shieldofststart+xshield, radiusinner)]))
        self.segments.append(Segment(  # guard (inner) bottom/left
            [(leadlen, -radiusinner), (leadlen+shieldofststart+xshield, -radiusinner)]))
        self.segments.append(Segment(  # shield (outer) top
            [(leadlen+shieldofststart, radiusouter),
             (length-leadlen-shieldofstend, radiusouter)]))
        self.segments.append(Segment(  # shield (outer) bottom
            [(leadlen+shieldofststart, -radiusouter),
             (length-leadlen-shieldofstend, -radiusouter)]))

        self.segments.append(SegmentArc(
            (leadlen, 0), width=radiusinner, height=radiusinner*2,
            theta1=0, theta2=360))
        self.segments.append(SegmentArc(
            (leadlen+shieldofststart, 0), width=radiusouter,
            height=radiusouter*2, theta1=-thetashield, theta2=thetashield))
        self.segments.append(SegmentArc(
            (length-leadlen-shieldofstend, 0), width=radiusouter,
            height=radiusouter*2, theta1=270, theta2=90))

        self.anchors['guardstart'] = (leadlen, -radiusinner)
        self.anchors['guardstart_top'] = (leadlen, radiusinner)
        self.anchors['guardend'] = (length-leadlen, -radiusinner)
        self.anchors['guardend_top'] = (length-leadlen, radiusinner)
        self.anchors['shieldstart'] = (leadlen+shieldofststart, -radiusouter)
        self.anchors['shieldstart_top'] = (leadlen+shieldofststart, radiusouter)
        self.anchors['shieldend'] = (length-leadlen-shieldofstend, -radiusouter)
        self.anchors['shieldend_top'] = (length-leadlen-shieldofstend, radiusouter)
        self.anchors['shieldcenter'] = (length/2, -radiusouter)
        self.anchors['shieldcenter_top'] = (length/2, radiusouter)

        if radiusouter <= radiusinner:
            warnings.warn('Triax outer radius < inner radius')

        if leadlen+shieldofststart > length-leadlen-shieldofstend:
            warnings.warn('Triax too short for outer radius')



''' Integrated Circuit Element '''

from __future__ import annotations

import sys
from typing import Optional, Sequence, cast
import math
from dataclasses import dataclass
from copy import copy

from ..segments import Segment, SegmentText, SegmentCircle, SegmentPoly, SegmentType
from ..elements import Element
from ..util import linspace, Point
from ..types import XY, Align, Side


@dataclass
class IcPin:
    ''' Integrated Circuit Pin

        Args:
            name: Input/output name (inside the box)
            pin: Pin name/number (outside the box)
            side: Side of box for the pin, 'left', 'right', 'top', 'bottom'
            pos: Pin position along the side, fraction from 0-1
            slot: Slot definition of pin location, given in 'X/Y' format.
                '2/4' is the second pin on a side with 4 pins.
            invert:Add an invert bubble to the pin
            invertradius: Radius of invert bubble
            color: Color for the pin and label
            rotation: Rotation for label text
            anchorname: Named anchor for the pin
     '''
    name: str | None = None
    pin: str | None = None
    side: Side = 'L'
    pos: float | None = None
    slot: str | None = None
    invert: bool = False
    invertradius: float = 0.15
    color: str | None = None
    rotation: float = 0
    anchorname: str | None = None
    lblsize: float | None = None


class Ic(Element):
    ''' Integrated Circuit element, or any other black-box element
        with arbitrary pins on any side.

        Args:
            size: (Width, Height) of box
            pins: List of IcPin instances defining the inputs/outputs
            pinspacing: Spacing between pins
            edgepadH: Padding between top/bottom and first pin
            edgepadW: Padding between left/right and first pin
            lofst: Offset between edge and label (inside box)
            lsize: Font size of labels (inside box)
            plblofst: Offset between edge and pin label (outside box)
            plblsize: Font size of pin labels (outside box)
            slant: Slant angle of top/bottom edges (e.g. for multiplexers)

        If a pin is named '>', it will be drawn as a proper clock signal input.

        Anchors:
            * inL[X] - Each pin on left side
            * inR[X] - Each pin on right side
            * inT[X] - Each pin on top side
            * inB[X] - Each pin on bottom side
            * pin[X] - Each pin with a number
            * CLK (if clock pin is defined with '>' name)

        Pins with names are also defined as anchors (if the name
        does not conflict with other attributes).
    '''
    def __init__(self,
                 size: XY=None,
                 pins: list[IcPin]=None,
                 pinspacing: float = 0.6,
                 edgepadH: float = 0.25,
                 edgepadW: float = 0.25,
                 leadlen: float = 0.5,
                 lofst: float = 0.15,
                 lsize: float = 14,
                 plblofst: float = 0.05,
                 plblsize: float = 11,
                 slant: float = 0,
                 w: float=None,
                 h: float=None,
                 **kwargs):
        super().__init__(**kwargs)

        # Sort pins by side
        if pins is None:
            pins = []
        pins = [copy(p) for p in pins]  # Make copy so user's IcPin instances don't change
        for pin in pins:
            # Convert pin designations to uppercase, single letter
            pin.side = pin.side[:1].upper()   # type: ignore
        sidepins = {}
        pincount = {}

        for side in ['L', 'R', 'T', 'B']:
            sidepins[side] = [p for p in pins if p.side == side]
            slotnames = [p.slot for p in sidepins[side]]
            # Add a 0 - can't max an empty list
            slots = [int(p.split('/')[1]) for p in slotnames if p is not None] + [0]
            pincount[side] = max(len(sidepins[side]), max(slots))

        if size is None:
            hcnt = max(pincount.get('L', 1), pincount.get('R', 1))
            wcnt = max(pincount.get('T', 1), pincount.get('B', 1))
            if h is None:
                try:
                    h = (hcnt-1)*pinspacing/(1-2/(hcnt+2)) + edgepadH*2
                except ZeroDivisionError:
                    h = 2 + edgepadH
            if w is None:
                try:
                    w = (wcnt-1)*pinspacing/(1-2/(wcnt+2)) + edgepadW*2
                except ZeroDivisionError:
                    w = 2 + edgepadW

            w = max(w, 2)  # Keep a minimum width for cases with 0-1 pins
            h = max(h, 2)
        else:
            w, h = size

        # Main box, adjusted for slant
        paths: list[list[Point]]
        if slant > 0:
            y1 = 0 - w * math.tan(math.radians(slant))
            y2 = h + w * math.tan(math.radians(slant))
            paths = [[Point((0, 0)), Point((w, y1)), Point((w, y2)), Point((0, h)), Point((0, 0))]]
        elif slant < 0:
            y1 = 0 + w * math.tan(math.radians(slant))
            y2 = h - w * math.tan(math.radians(slant))
            paths = [[Point((0, y1)), Point((w, 0)), Point((w, h)), Point((0, y2)), Point((0, y1))]]
        else:
            y1 = 0
            y2 = h
            paths = [[Point((0, 0)), Point((w, 0)), Point((w, h)), Point((0, h)), Point((0, 0))]]

        # Add each pin
        for side in sidepins.keys():
            if side in ['L', 'R']:
                sidelen = h-edgepadH*2
            else:
                sidelen = w-edgepadW*2

            leadext = {'L': Point((-leadlen, 0)),
                       'R': Point((leadlen, 0)),
                       'T': Point((0, leadlen)),
                       'B': Point((0, -leadlen))}.get(side, Point((0,0)))

            for i, pin in enumerate(sidepins[side]):
                # Determine pin position
                if pin.pos is not None:
                    z = pin.pos * sidelen
                elif pin.slot is None:
                    pin.slot = '{}/{}'.format(i+1, len(sidepins[side]))

                if pin.slot is not None:
                    num, tot = [int(k) for k in pin.slot.split('/')]
                    if tot == 1:
                        z = sidelen/2  # Single pin, center it
                    else:
                        # Evenly spaced along side
                        z = linspace(1/(tot+2), 1-1/(tot+2), num=tot)[num-1] * sidelen

                pinxy = {'L': Point((0, z+edgepadH)),
                         'R': Point((w, z+edgepadH)),
                         'T': Point((z+edgepadW, h)),
                         'B': Point((z+edgepadW, 0))}.get(side, Point((0,0)))

                # Adjust pin position for slant
                if side == 'T' and slant > 0:
                    pinxy = Point((pinxy[0], pinxy[1] - pinxy[0] * math.tan(-math.radians(slant))))
                elif side == 'T' and slant < 0:
                    pinxy = Point(((pinxy[0], pinxy[1] + (y2-h) - pinxy[0] * math.tan(-math.radians(slant)))))
                elif side == 'B' and slant < 0:
                    pinxy = Point((pinxy[0], pinxy[1] - (y2-h) - pinxy[0] * math.tan(math.radians(slant))))
                elif side == 'B' and slant > 0:
                    pinxy = Point((pinxy[0], pinxy[1] - pinxy[0] * math.tan(math.radians(slant))))

                if pin.name == '>':
                    # Draw clock pin
                    clkxy = pinxy
                    clkw, clkh = 0.4 * lsize/16, 0.2 * lsize/16
                    if side in ['T', 'B']:
                        clkw = math.copysign(clkw, leadext[1]) if leadext[1] != 0 else clkw
                        clkpath = [Point((clkxy[0]-clkh, clkxy[1])),
                                   Point((clkxy[0], clkxy[1]-clkw)),
                                   Point((clkxy[0]+clkh, clkxy[1]))]
                    else:
                        clkw = math.copysign(clkw, -leadext[0]) if leadext[0] != 0 else clkw
                        clkpath = [Point((clkxy[0], clkxy[1]+clkh)),
                                   Point((clkxy[0]+clkw, clkxy[1])),
                                   Point((clkxy[0], clkxy[1]-clkh))]
                    paths.append(clkpath)

                elif pin.name and pin.name != '':
                    # Add pin label
                    pofst = {'L': Point((lofst, 0)),
                             'R': Point((-lofst, 0)),
                             'T': Point((0, -lofst)),
                             'B': Point((0, lofst))}.get(side)

                    align = cast(Optional[Align], {'L': ('left', 'center'),
                             'R': ('right', 'center'),
                             'T': ('center', 'top'),
                             'B': ('center', 'bottom')}.get(side))

                    self.segments.append(SegmentText(pos=pinxy+pofst,
                                         label=pin.name,
                                         align=align,
                                         fontsize=pin.lblsize if pin.lblsize is not None else plblsize,
                                         color=pin.color,
                                         rotation=pin.rotation,
                                         rotation_mode='default')
                                        )

                # Add pin number outside the IC
                if pin.pin and pin.pin != '':
                    # Account for any invert-bubbles
                    invertradius = pin.invertradius * pin.invert
                    pofst = {'L': Point((-plblofst-invertradius*2, plblofst)),
                             'R': Point((plblofst+invertradius*2, plblofst)),
                             'T': Point((plblofst, plblofst+invertradius*2)),
                             'B': Point((plblofst, -plblofst-invertradius*2))
                             }.get(side)

                    align = cast(Optional[Align], {'L': ('right', 'bottom'),
                                                   'R': ('left', 'bottom'),
                                                   'T': ('left', 'bottom'),
                                                   'B': ('left', 'top')}.get(side))
                    self.segments.append(SegmentText(pos=pinxy+pofst,
                                                     label=pin.pin,
                                                     align=align,
                                                     fontsize=pin.lblsize if pin.lblsize is not None else plblsize
                                                    ))

                # Draw leads
                if leadlen > 0:
                    if pin.invert:
                        # Add invert-bubble
                        invertradius = pin.invertradius
                        invertofst = {'L': Point((-invertradius, 0)),
                                      'R': Point((invertradius, 0)),
                                      'T': Point((0, invertradius)),
                                      'B': Point((0, -invertradius))}.get(side, Point((0,0)))

                        self.segments.append(SegmentCircle(
                            pinxy+invertofst, invertradius))
                        paths.append([pinxy+invertofst*2, pinxy+leadext])
                    else:
                        paths.append([pinxy, pinxy+leadext])

                # Define anchors
                anchorpos = pinxy+leadext
                self.anchors['in{}{}'.format(side[0].upper(), i+1)] = anchorpos
                if pin.anchorname:
                    self.anchors[pin.anchorname] = anchorpos
                elif pin.name:
                    if pin.name == '>':
                        self.anchors['CLK'] = anchorpos
                    self.anchors[pin.name] = anchorpos
                if pin.pin:
                    self.anchors['pin{}'.format(pin.pin)] = anchorpos

        for p in paths:
            self.segments.append(Segment(p))
        self.params['lblloc'] = 'center'
        self.anchors['center'] = (w/2, h/2)


class Multiplexer(Ic):
    ''' Multiplexer

        Args:
            demux: Draw as demultiplexer
            size: (Width, Height) of box
            pins: List of IcPin instances defining the inputs/outputs
            pinspacing: Spacing between pins
            edgepadH: Padding between top/bottom and first pin
            edgepadW: Padding between left/right and first pin
            lofst: Offset between edge and label (inside box)
            lsize: Font size of labels (inside box)
            plblofst: Offset between edge and pin label (outside box)
            plblsize: Font size of pin labels (outside box)
            slant: Slant angle of top/bottom edges

        If a pin is named '>', it will be drawn as a proper clock signal input.

        Anchors:
            * inL[X] - Each pin on left side
            * inR[X] - Each pin on right side
            * inT[X] - Each pin on top side
            * inB[X] - Each pin on bottom side
            * pin[X] - Each pin with a number
            * CLK (if clock pin is defined with '>' name)

        Pins with names are also defined as anchors (if the name does
        not conflict with other attributes).
    '''
    def __init__(self,
                 demux: bool = False,
                 size: XY=None,
                 pins: list[IcPin]=None,
                 pinspacing: float = 0.6,
                 edgepadH: float = 0.25,
                 edgepadW: float = 0.25,
                 leadlen: float = 0.5,
                 lofst: float = 0.15,
                 lsize: float = 14,
                 plblofst: float = 0.05,
                 plblsize: float = 11,
                 slant: float = 25,
                 **kwargs):
        slant = -slant if not demux else slant
        super().__init__(size=size,
                         pins=pins,
                         pinspacing=pinspacing,
                         edgepadH=edgepadH,
                         edgepadW=edgepadW,
                         leadlen=leadlen,
                         lofst=lofst,
                         lsize=lsize,
                         plblofst=plblofst,
                         plblsize=plblsize,
                         slant=slant,
                         **kwargs)


class IcDIP(Element):
    ''' Dual-inline Package Integrated Circuit.

        Args:
            pins: number of pins
            names: List of names for each pin to display inside the box
            notch: Show the notch at top of box
            width: Width of the box
            pinw: Width and height of each pin
            spacing: Distance between each pin
            number: Show pin numbers inside each pin
            fontsize: Size for pin name labels
            pfontsize: Size for pin number labels

        Anchors:
            * p[x]  - Each pin
            * p[x]_in  - Inside contact for each pin

        If signal names are provided, they will also be added as anchors
        along with _in inside variants.
    '''
    def __init__(self, *d, pins: int=8, 
                 names: Sequence[str]=None,
                 notch: bool=True,
                 width: float=3,
                 pinw: float=0.6,
                 spacing: float=0.5,
                 number: bool=True,
                 fontsize: float=12,
                 pfontsize: float=10,         
                 **kwargs):
        super().__init__(*d, **kwargs)
        if pins % 2 == 1:
            raise ValueError('pins must be even')

        height = pins/2*pinw + spacing*(pins/2+1)    
        self.segments.append(SegmentPoly(((0, 0), (width, 0), (width, height), (0, height))))
        
        if notch:
            c = width/2
            theta = linspace(-math.pi, 0)
            notchr = pinw/2
            notchx = [c + notchr * math.cos(t) for t in theta]
            notchy = [height + notchr * math.sin(t) for t in theta]
            self.segments.append(Segment(list(zip(notchx, notchy))))
        
        for i in range(pins//2):
            y1 = spacing + spacing*i+i*pinw
            y2 = spacing + spacing*i+(i+1)*pinw
            ymid = (y1+y2)/2

            # Left
            self.segments.append(SegmentPoly(((0, y1), (0, y2), (-pinw, y2), (-pinw, y1))))
            pnum = pins//2 - i
            self.anchors[f'p{pnum}'] = (-pinw, ymid)
            self.anchors[f'p{pnum}_in'] = (0, ymid)
            if number:
                self.segments.append(SegmentText((-pinw/2, ymid), str(pnum), fontsize=pfontsize))
            if names:
                self.segments.append(SegmentText((.1, ymid), names[pnum-1], align=('left', 'center'), fontsize=fontsize))
                self.anchors[names[pnum-1]] = (-pinw, ymid)
                self.anchors[f'{names[pnum-1]}_in'] = (0, ymid)
                
            # Right
            self.segments.append(SegmentPoly(((width, y1), (width, y2), (width+pinw, y2), (width+pinw, y1))))
            pnum = pins//2 + i + 1
            self.anchors[f'p{pnum}'] = (width+pinw, ymid)
            self.anchors[f'p{pnum}_in'] = (width, ymid)
            if number:
                self.segments.append(SegmentText((width+pinw/2, ymid), str(pnum), fontsize=pfontsize))
            if names:
                self.segments.append(SegmentText((width-.1, ymid), names[pnum-1], align=('right', 'center'), fontsize=fontsize))
                self.anchors[names[pnum-1]] = (width+pinw, ymid)
                self.anchors[f'{names[pnum-1]}_in'] = (width, ymid)


class DFlipFlop(Ic):
    ''' D-Type Flip Flop

        Args:
            preclr: Show preset and clear inputs
            preclrinvert: Add invert bubble to preset and clear inputs
            size: Size of the box

        Anchors:
            * D
            * CLK
            * Q
            * Qbar
            * PRE
            * CLR
    '''
    def __init__(self, *d, preclr: bool=False, preclrinvert: bool=True, size=(2, 3), **kwargs):
        pins=[IcPin('D', side='left', slot='2/2'),
              IcPin('>', side='left', slot='1/2'),
              IcPin('Q', side='right', slot='2/2'),
              IcPin('$\overline{\mathrm{Q}}$', side='right', slot='1/2', anchorname='Qbar')]

        if preclr:
            pins.extend([IcPin('PRE', side='top', invert=preclrinvert),
                         IcPin('CLR', side='bottom', invert=preclrinvert)])

        super().__init__(pins=pins, size=size)
        

class JKFlipFlop(Ic):
    ''' J-K Flip Flop

        Args:
            preclr: Show preset and clear inputs
            preclrinvert: Add invert bubble to preset and clear inputs
            size: Size of the box

        Anchors:
            * J
            * K
            * CLK
            * Q
            * Qbar
            * PRE
            * CLR
    '''
    def __init__(self, *d, preclr: bool=False, preclrinvert: bool=True, size=(2, 3), **kwargs):
        pins=[IcPin('J', side='left', slot='3/3'),
              IcPin('>', side='left', slot='2/3'),
              IcPin('K', side='left', slot='1/3'),
              IcPin('Q', side='right', slot='2/2'),
              IcPin('$\overline{\mathrm{Q}}$', side='right', slot='1/2', anchorname='Qbar')]

        if preclr:
            pins.extend([IcPin('PRE', side='top', invert=preclrinvert),
                         IcPin('CLR', side='bottom', invert=preclrinvert)])

        super().__init__(pins=pins, size=size)


class VoltageRegulator(Ic):
    ''' Voltage regulator

        Args:
            size: Size of the box

        Anchors:
            * in
            * out
            * gnd
    '''
    def __init__(self, *d, size=(2, 1.5), **kwargs):
        pins=[IcPin('in', side='left', slot='3/3'),
              IcPin('out', side='right', slot='3/3'),
              IcPin('gnd', side='bottom')]
        super().__init__(pins=pins, size=size)


class Ic555(Ic):
    def __init__(self, *d, **kwargs):
        pins=[IcPin(name='TRG', side='left', pin='2'),
              IcPin(name='THR', side='left', pin='6'),
              IcPin(name='DIS', side='left', pin='7'),
              IcPin(name='CTL', side='right', pin='5'),
              IcPin(name='OUT', side='right', pin='3'),
              IcPin(name='RST', side='top', pin='4'),
              IcPin(name='Vcc', side='top', pin='8'),
              IcPin(name='GND', side='bot', pin='1')]
        super().__init__(pins=pins,
                         edgepadW=.5,
                         edgepadH=1,
                         pinspacing=1.5,
                         leadlen=1,
                         label='555')
        
        
def sevensegdigit(bottom: float=0, left: float=0,
                  seglen: float=1.5, segw: float=0.3, spacing: float=0.12,                  
                  decimal: bool=False, digit: int | str=8, 
                  segcolor: str='red', tilt: float=10, labelsegments: bool=True) -> list[SegmentType]:
    ''' Generate drawing segments for a 7-segment display digit. Use for
        building new elements incorporating a 7-segment display.
    
        Args:
            bottom: Location of bottom of digit
            left: Location of left side of digit
            seglen: Length of one segment
            segw: Width of one segment
            spacing: Distance between segments in corners
            decimal: Show decimal point segment
            digit: Number to display
            segcolor: Color of segments
            tilt: Tilt angle in degrees
            labelsegments: Add a-g labels to each segment
            anode: Add common anode pin
            cathode: Add common cathode pin
            size: Total size of the box

        Returns:
            List of Segments making the digit
    '''
    halfw = segw/2 # Half segment width
    halfspace = spacing/2
    tilt = math.radians(tilt)

    # Straight (non-tilted) segments - Horizontal
    segDx = [halfspace, halfspace+halfw, seglen-halfspace-halfw,
             seglen-halfspace, seglen-halfspace-halfw, halfspace+halfw]
    segDy = [0, halfw, halfw, 0, -halfw, -halfw]
    segAx = segDx
    segGx = segDx
    segAy = [k + seglen*2 for k in segDy]
    segGy = [k + seglen for k in segDy]
    # - Vertical
    segEx = [0, halfw, halfw, 0, -halfw, -halfw]
    segEy = [halfspace, halfspace+halfw, seglen-halfspace-halfw,
             seglen-halfspace,  seglen-halfspace-halfw, halfspace+halfw]
    segFx = segEx
    segFy = [k + seglen for k in segEy]
    segBx = [k + seglen for k in segEx]
    segBy = segFy
    segCx = segBx
    segCy = segEy

    # Label positions
    segAlabel = (seglen/2 + halfspace, seglen*2+halfw)
    segGlabel = (seglen/2 + halfspace, seglen+halfw)
    segDlabel = (seglen/2 + halfspace, halfw)
    segBlabel = (seglen + halfw + halfspace, seglen*1.5)
    segClabel = (seglen + halfw + halfspace, seglen*.5)
    segElabel = (-halfw - halfspace, seglen*.5)
    segFlabel = (-halfw - halfspace, seglen*1.5)

    # Apply shear
    if tilt != 0:
        lam = math.sin(tilt)
        segAx = [k + lam*i for k, i in zip(segAx, segAy)]
        segBx = [k + lam*i for k, i in zip(segBx, segBy)]
        segCx = [k + lam*i for k, i in zip(segCx, segCy)]
        segDx = [k + lam*i for k, i in zip(segDx, segDy)]
        segEx = [k + lam*i for k, i in zip(segEx, segEy)]
        segFx = [k + lam*i for k, i in zip(segFx, segFy)]
        segGx = [k + lam*i for k, i in zip(segGx, segGy)]
        def shiftlabel(label):
            return label[0] + label[1]*lam, label[1]
        segAlabel = shiftlabel(segAlabel)
        segGlabel = shiftlabel(segGlabel)
        segDlabel = shiftlabel(segDlabel)
        segBlabel = shiftlabel(segBlabel)
        segClabel = shiftlabel(segClabel)
        segElabel = shiftlabel(segElabel)
        segFlabel = shiftlabel(segFlabel)
        left -= lam*seglen    

    # Translate to final position
    segDx = [k + left for k in segDx]
    segDy = [k + bottom for k in segDy]
    segGx = [k + left for k in segGx]
    segGy = [k + bottom for k in segGy]
    segAx = [k + left for k in segAx]
    segAy = [k + bottom for k in segAy]
    segEx = [k + left for k in segEx]
    segEy = [k + bottom for k in segEy]
    segFx = [k + left for k in segFx]
    segFy = [k + bottom for k in segFy]
    segBx = [k + left for k in segBx]
    segBy = [k + bottom for k in segBy]
    segCx = [k + left for k in segCx]
    segCy = [k + bottom for k in segCy]
    segAlabel = (segAlabel[0] + left, segAlabel[1] + bottom)
    segGlabel = (segGlabel[0] + left, segGlabel[1] + bottom)
    segDlabel = (segDlabel[0] + left, segDlabel[1] + bottom)
    segBlabel = (segBlabel[0] + left, segBlabel[1] + bottom)
    segClabel = (segClabel[0] + left, segClabel[1] + bottom)
    segElabel = (segElabel[0] + left, segElabel[1] + bottom)
    segFlabel = (segFlabel[0] + left, segFlabel[1] + bottom)

    # Fill based on digit parameter
    fillA = segcolor if str(digit).lower() in ['2', '3', '5', '6', '7', '8', '9', '0', 'a', 'c', 'e', 'f'] else None
    fillB = segcolor if str(digit).lower() in ['1', '2', '3', '4', '7', '8', '9', '0', 'a', 'd'] else None
    fillC = segcolor if str(digit).lower() in ['1', '3', '4', '5', '6', '7', '8', '9', '0', 'a', 'b', 'd'] else None
    fillD = segcolor if str(digit).lower() in ['2', '3', '5', '6', '8', '9', '0', 'b', 'c', 'd', 'e'] else None
    fillE = segcolor if str(digit).lower() in ['2', '6', '8', '0', 'a', 'b', 'c', 'd', 'e', 'f'] else None
    fillF = segcolor if str(digit).lower() in ['4', '5', '6', '8', '9', '0', 'a', 'b', 'c', 'e', 'f'] else None
    fillG = segcolor if str(digit).lower() in ['2', '3', '4', '5', '6', '8', '9', 'a', 'b', 'd', 'e', 'f'] else None

    segments: list[SegmentType] = []
    segments.append(SegmentPoly(list(zip(segAx, segAy)), color='gray', fill=fillA, lw=.5))
    segments.append(SegmentPoly(list(zip(segBx, segBy)), color='gray', fill=fillB, lw=.5))
    segments.append(SegmentPoly(list(zip(segCx, segCy)), color='gray', fill=fillC, lw=.5))
    segments.append(SegmentPoly(list(zip(segDx, segDy)), color='gray', fill=fillD, lw=.5))
    segments.append(SegmentPoly(list(zip(segEx, segEy)), color='gray', fill=fillE, lw=.5))
    segments.append(SegmentPoly(list(zip(segFx, segFy)), color='gray', fill=fillF, lw=.5))
    segments.append(SegmentPoly(list(zip(segGx, segGy)), color='gray', fill=fillG, lw=.5))
    if labelsegments:
        segments.append(SegmentText(segAlabel, 'a', align=('center', 'bottom'), fontsize=10))
        segments.append(SegmentText(segGlabel, 'g', align=('center', 'bottom'), fontsize=10))
        segments.append(SegmentText(segDlabel, 'd', align=('center', 'bottom'), fontsize=10))
        segments.append(SegmentText(segBlabel, 'b', align=('left', 'center'), fontsize=10))
        segments.append(SegmentText(segClabel, 'c', align=('left', 'center'), fontsize=10))
        segments.append(SegmentText(segElabel, 'e', align=('right', 'center'), fontsize=10))
        segments.append(SegmentText(segFlabel, 'f', align=('right', 'center'), fontsize=10))

    if decimal:
        dotrad = .15
        segments.append(SegmentCircle((left+seglen+segw+dotrad/2, bottom),
                                      radius=dotrad, color='gray', fill=segcolor, lw=.5))
    return segments
        
        
class SevenSegment(Ic):
    ''' A seven-segment display digit.

        Args:
            decimal: Show decimal point segment
            digit: Number to display
            segcolor: Color of segments
            tilt: Tilt angle in degrees
            labelsegments: Add a-g labels to each segment
            anode: Add common anode pin
            cathode: Add common cathode pin
            size: Total size of the box

        Anchors:
            * a
            * b
            * c
            * d
            * e
            * f
            * g
            * dp
            * cathode
            * anode
    '''
    def __init__(self, *d, decimal: bool=False,
                 digit: int | str=8, 
                 segcolor: str='red',
                 tilt: float=10,
                 labelsegments: bool=True,
                 anode: bool=False,
                 cathode: bool=False,
                 size=(2, 1.5), **kwargs):
        
        if decimal:
            slots = '8'
            boxheight = 5.9
        else:
            slots = '7'
            boxheight = 5.3
        
        pins=[IcPin(pin='a', side='left', slot=f'{7+decimal}/{slots}', anchorname='a'),
              IcPin(pin='b', side='left', slot=f'{6+decimal}/{slots}', anchorname='b'),
              IcPin(pin='c', side='left', slot=f'{5+decimal}/{slots}', anchorname='c'),
              IcPin(pin='d', side='left', slot=f'{4+decimal}/{slots}', anchorname='d'),
              IcPin(pin='e', side='left', slot=f'{3+decimal}/{slots}', anchorname='e'),
              IcPin(pin='f', side='left', slot=f'{2+decimal}/{slots}', anchorname='f'),
              IcPin(pin='g', side='left', slot=f'{1+decimal}/{slots}', anchorname='g')]
        if decimal:
            pins.append(IcPin(pin='dp', side='left', slot=f'1/{slots}', anchorname='dp'))
        if anode:
            pins.append(IcPin(pin='ca', side='top', anchorname='anode'))
        if cathode:
            pins.append(IcPin(pin='cc', side='bottom', anchorname='cathode'))

        super().__init__(pins=pins, w=3)

        left = 0.8
        seglen = 1.5
        bot = (boxheight-seglen*2)/2
        
        segments = sevensegdigit(left=0.8, bottom=bot, decimal=decimal, digit=digit,
                                 segcolor=segcolor, tilt=tilt, labelsegments=labelsegments)
        self.segments.extend(segments)


from .process import Stirrer, Cistern, Tank, HeatExchanger, PlateExchanger, DiscCentrifuge, Pump, Valve, Boiler, CoolingTower, Filter, Sieve
from ..dsp import Square, Circle, Sum, Mixer
from ..elements import Arrowhead, Dot, Arrow, Line, LineDot, Ic, IcPin, Multiplexer
from ..flow import Box

import warnings



''' Other elements '''

from .elements import Element, Element2Term, gap
from .twoterm import resheight
from ..segments import Segment, SegmentPoly, SegmentArc, SegmentCircle, SegmentArrow


class Speaker(Element):
    ''' Speaker element with two inputs.

        Anchors:
            * in1
            * in2
    '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        sph = .5
        self.segments.append(Segment([(0, 0), (resheight, 0)]))
        self.segments.append(Segment([(0, -sph), (resheight, -sph)]))
        self.segments.append(SegmentPoly(
            [(resheight, sph/2), (resheight, -sph*1.5),
             (resheight*2, -sph*1.5), (resheight*2, sph/2)]))
        self.segments.append(SegmentPoly(
            [(resheight*2, sph/2), (resheight*3.5, sph*1.25),
             (resheight*3.5, -sph*2.25), (resheight*2, -sph*1.5)],
            closed=False))
        self.anchors['in1'] = (0, 0)
        self.anchors['in2'] = (0, -sph)
        self.params['drop'] = (0, -sph)


class Mic(Element):
    ''' Microphone element with two inputs.

        Anchors:
            * in1
            * in2
    '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        sph = .5
        self.segments.append(Segment(  # Upper lead
            [(0, 0), (resheight, 0)]))
        self.segments.append(Segment(  # Lower lead
            [(0, -sph), (resheight, -sph)]))
        self.segments.append(Segment(  # Vertical flat
            [(-resheight*2, resheight), (-resheight*2, -resheight*3)]))
        self.segments.append(SegmentArc(
            (-resheight*2, -resheight), theta1=270, theta2=90,
            width=resheight*4, height=resheight*4))
        self.anchors['in1'] = (resheight, 0)
        self.anchors['in2'] = (resheight, -sph)
        self.params['drop'] = (0, -sph)


class Motor(Element2Term):
    ''' Motor '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        mw = .22
        self.segments.append(Segment(
            [(-mw, 0), (-mw, 0), gap, (1+mw, 0), (1+mw, 0)]))
        self.segments.append(Segment(
            [(0, -mw), (0-mw, -mw), (0-mw, mw), (0, mw)]))
        self.segments.append(Segment(
            [(1, -mw), (1+mw, -mw), (1+mw, mw), (1, mw)]))
        self.segments.append(SegmentCircle((0.5, 0), 0.5))


class AudioJack(Element):
    ''' Audio Jack with 2 or 3 connectors and optional switches.

        Args:
            ring: Show ring (third conductor) contact
            switch: Show switch on tip contact
            ringswitch: Show switch on ring contact
            dots: Show connector dots
            radius: Radius of connector dots

        Anchors:
            * tip
            * sleeve
            * ring
            * ringswitch
            * tipswitch
    '''
    def __init__(self, *d,
                 radius: float=0.075,
                 ring: bool=False,
                 ringswitch: bool=False,
                 dots: bool=True,
                 switch: bool=False,
                 open: bool=True,
                 **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'bg' if open else None

        length = 2.0
        ringlen = .75
        tiplen = .55
        swidth = .2
        sleeveheight = 1.0
        tipy = 1.0
        ringy = .1
        sleevey = .35
        swdy = .4
        swlen = .5

        if switch:
            tipy += .2

        if ring and ringswitch:
            sleevey += .2
            ringy -= .2

        if ring:
            if dots:
                self.segments.append(SegmentCircle((0, -sleevey), radius,
                                                   fill=fill, zorder=4))
            self.segments.append(Segment(
                [(-radius, -sleevey), (-length, -sleevey),
                 (-length, 0), (-length, sleeveheight),
                 (-length-swidth, sleeveheight),
                 (-length-swidth, 0), (-length, 0)]))
            self.anchors['sleeve'] = (0, -sleevey)

            if dots:
                self.segments.append(SegmentCircle(
                    (0,  ringy), radius, fill=fill, zorder=4))
            self.segments.append(Segment(
                [(-radius, ringy), (-length*.75, ringy),
                 (-length*ringlen-2*radius, ringy+2*radius),
                 (-length*ringlen-radius*4, ringy)]))
            self.anchors['ring'] = (0, ringy)

        else:
            if dots:
                self.segments.append(SegmentCircle(
                    (0, 0), radius, fill=fill, zorder=4))
            self.segments.append(Segment(
                [(-radius, 0), (-length, 0), (-length, sleeveheight),
                 (-length+swidth, sleeveheight), (-length+swidth, 0)]))
            self.anchors['sleeve'] = (0, 0)

        if dots:
            self.segments.append(SegmentCircle(
                (0, tipy), radius, fill=fill, zorder=4))
        self.segments.append(Segment(
            [(-radius, tipy), (-length*.55, tipy),
             (-length*tiplen-2*radius, tipy-2*radius),
             (-length*tiplen-radius*4, tipy)]))
        self.anchors['tip'] = (0, tipy)

        if switch:
            if dots:
                self.segments.append(SegmentCircle(
                    (0, tipy-swdy), radius, fill=fill, zorder=4))
            self.segments.append(Segment(
                [(0, tipy-swdy), (-swlen, tipy-swdy)]))
            self.segments.append(SegmentArrow(
                (-swlen, tipy-swdy), (-swlen, tipy)))
            self.anchors['tipswitch'] = (0, tipy-swdy)

        if ring and ringswitch:
            if dots:
                self.segments.append(SegmentCircle(
                    (0, ringy+swdy), radius, fill=fill, zorder=4))
            self.segments.append(Segment(
                [(0, ringy+swdy), (-swlen, ringy+swdy)]))
            self.segments.append(SegmentArrow(
                (-swlen, ringy+swdy), (-swlen, ringy)))
            self.anchors['ringswitch'] = (0, ringy+swdy)



''' Operation amplifier '''

import math

from .elements import Element, gap
from ..segments import Segment


oa_back = 2.5
oa_xlen = oa_back * math.sqrt(3)/2
oa_lblx = oa_xlen/8
oa_pluslen = .2


class Opamp(Element):
    ''' Operational Amplifier.

        Args:
            sign: Draw +/- labels at each input

        Anchors:
            * in1
            * in2
            * out
            * vd
            * vs
            * n1
            * n2
            * n1a
            * n2a
    '''
    def __init__(self, *d, sign: bool=True, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), (0, oa_back/2), (oa_xlen, 0), (0, -oa_back/2), (0, 0),
             gap, (oa_xlen, 0)]))

        if sign:
            self.segments.append(Segment(
                [(oa_lblx-oa_pluslen/2, oa_back/4),
                 (oa_lblx+oa_pluslen/2, oa_back/4)]))
            self.segments.append(Segment(
                [(oa_lblx-oa_pluslen/2, -oa_back/4),
                 (oa_lblx+oa_pluslen/2, -oa_back/4)]))
            self.segments.append(Segment(
                [(oa_lblx, -oa_back/4-oa_pluslen/2),
                 (oa_lblx, -oa_back/4+oa_pluslen/2)]))

        self.anchors['center'] = (oa_xlen/2, 0)
        self.anchors['in1'] = (0, oa_back/4)
        self.anchors['in2'] = (0, -oa_back/4)
        self.anchors['out'] = (oa_xlen, 0)
        self.anchors['vd'] = (oa_xlen/3, .84)
        self.anchors['vs'] = (oa_xlen/3, -.84)
        self.anchors['n1'] = (oa_xlen*2/3, -.42)
        self.anchors['n2'] = (oa_xlen*2/3, .42)
        self.anchors['n1a'] = (oa_xlen*.9, -.13)
        self.anchors['n2a'] = (oa_xlen*.9, .13)
        self.params['drop'] = (oa_xlen, 0)



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name = 'schemdraw',
    version = '0.12',
    description = 'Electrical circuit schematic drawing',
    author = 'Collin J. Delker',
    author_email = 'schemdraw@collindelker.com',
    url = 'https://schemdraw.readthedocs.io/',
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        'Source': 'https://bitbucket.org/cdelker/schemdraw',
    },
    python_requires='>=3.7',
    packages = ['schemdraw',
                'schemdraw.elements',
                'schemdraw.logic',
                'schemdraw.dsp',
                'schemdraw.flow',
                'schemdraw.backends',
                'schemdraw.parsing'
               ],
    package_dir={'schemdraw': 'schemdraw'},
    keywords = ['circuit', 'schematic', 'electrical', 'flowchart', 'logic'],
    install_requires=['typing_extensions; python_version<"3.8"'],
    extras_require={
        'matplotlib':  ['matplotlib>=3'],
        'svgmath': ['ziafont>=0.2', 'ziamath>=0.3', 'latex2mathml']
    },
    package_data = {'schemdraw': ['py.typed']},
    zip_safe=False,
    classifiers = [
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'Intended Audience :: End Users/Desktop',
    ],
)



''' Flowcharting element definitions '''

import math

from ..util import linspace
from ..segments import Segment, SegmentCircle, SegmentText
from ..elements import Element


class Box(Element):
    ''' Flowchart box

        Args:
            w: Width of box
            h: Height of box

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, w: float=3, h: float=2, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, w/2), (h, w/2),
                                      (h, -w/2), (0, -w/2), (0, 0)]))
        self.params['lblloc'] = 'center'
        self.params['lblofst'] = 0
        self.params['theta'] = -90
        self.params['drop'] = (h, 0)
        self.anchors['W'] = (h/2, -w/2)
        self.anchors['E'] = (h/2, w/2)
        self.anchors['S'] = (h, 0)
        self.anchors['N'] = (0, 0)


class Subroutine(Element):
    ''' Flowchart subroutine. Box with extra vertical lines
        near sides.

        Args:
            w: Width of box
            h: Height of box
            s: spacing of side lines

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, w: float=3.5, h: float=2, s: float=0.3, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, w/2), (h, w/2),
                                      (h, -w/2), (0, -w/2), (0, 0)]))
        self.segments.append(Segment([(0, w/2-s), (h, w/2-s)]))
        self.segments.append(Segment([(0, -w/2+s), (h, -w/2+s)]))
        self.params['lblloc'] = 'center'
        self.params['lblofst'] = 0
        self.params['theta'] = -90
        self.params['drop'] = (h, 0)
        self.anchors['W'] = (h/2, -w/2)
        self.anchors['E'] = (h/2, w/2)
        self.anchors['S'] = (h, 0)
        self.anchors['N'] = (0, 0)


class Data(Element):
    ''' Flowchart data box (parallelogram)

        Args:
            w: Width of box
            h: Height of box
            s: slant of sides

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, w: float=3, h: float=2, s: float=0.5, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, w/2+s/2), (h, w/2-s/2),
                                      (h, -w/2-s/2), (0, -w/2+s/2), (0, 0)]))
        self.params['lblloc'] = 'center'
        self.params['lblofst'] = 0
        self.params['theta'] = -90
        self.params['drop'] = (h, 0)
        self.anchors['W'] = (h/2, -w/2)
        self.anchors['E'] = (h/2, w/2)
        self.anchors['S'] = (h, 0)
        self.anchors['N'] = (0, 0)


class Start(Element):
    ''' Flowchart start/stop box (ellipse)


        Args:
            w: Width of ellipse
            h: Height of ellipse

        Anchors:
            * N
            * S
            * E
            * W    '''
    def __init__(self, w: float=3, h: float=2, **kwargs):
        super().__init__(**kwargs)
        # There's no ellipse Segment type, so draw one with a path Segment
        t = linspace(0, math.pi*2, num=50)
        y = [(w/2) * math.cos(t0) for t0 in t]
        x = [(h/2) * math.sin(t0) + h/2 for t0 in t]
        x[-1] = x[0]
        y[-1] = y[0]  # Ensure the path is actually closed
        self.segments.append(Segment(list(zip(x, y))))
        self.params['lblloc'] = 'center'
        self.params['lblofst'] = 0
        self.params['theta'] = -90
        self.params['drop'] = (h, 0)
        self.anchors['W'] = (h/2, -w/2)
        self.anchors['E'] = (h/2, w/2)
        self.anchors['S'] = (h, 0)
        self.anchors['N'] = (0, 0)


class Decision(Element):
    ''' Flowchart decision (diamond)


        Args:
            w: Width of box
            h: Height of box
            N: text for North decision branch
            S: text for South decision branch
            E: text for East decision branch
            W: text for West decision branch
            font: Font family/name
            fontsize: Point size of label font

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, w: float=4, h: float=2,
                 N: str=None, E: str=None, S: str=None, W: str=None,
                 font: str=None, fontsize: float=14,
                 **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (h/2, w/2), (h, 0),
                                      (h/2, -w/2), (0, 0)]))
        self.params['lblloc'] = 'center'
        self.params['lblofst'] = 0
        self.params['theta'] = -90
        self.params['drop'] = (h, 0)
        self.anchors['W'] = (h/2, -w/2)
        self.anchors['E'] = (h/2, w/2)
        self.anchors['S'] = (h, 0)
        self.anchors['N'] = (0, 0)

        lblofst = .13
        if N:
            self.segments.append(SegmentText(
                (0, lblofst), N, align=('left', 'bottom'),
                font=font, fontsize=fontsize))
        if S:
            self.segments.append(SegmentText(
                (h, lblofst), S, align=('left', 'top'),
                font=font, fontsize=fontsize))
        if E:
            self.segments.append(SegmentText(
                (h/2, w/2+lblofst), E, align=('left', 'bottom'),
                font=font, fontsize=fontsize))
        if W:
            self.segments.append(SegmentText(
                (h/2, -w/2-lblofst), W, align=('right', 'bottom'),
                font=font, fontsize=fontsize))


class Connect(Element):
    ''' Flowchart connector (circle)

        Args:
            r: Radius of box

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        r = kwargs.get('r', 0.75)
        self.segments.append(SegmentCircle((r, 0), r))
        self.params['lblloc'] = 'center'
        self.params['lblofst'] = 0
        self.params['theta'] = -90
        self.params['drop'] = (2*r, 0)
        self.anchors['W'] = (r, -r)
        self.anchors['E'] = (r, r)
        self.anchors['S'] = (2*r, 0)
        self.anchors['N'] = (0, 0)



''' Test for using schemdraw in a script WITHOUT interactive pyplot window  '''
import schemdraw
import schemdraw.elements as elm

schemdraw.use('svg')

d = schemdraw.Drawing()
d.add(elm.Resistor().label('1K'))
d.add(elm.Capacitor().down())
d.save('testcircuit.png')
print(d.get_imagedata('svg'))



''' Schemdraw drawing segments.

    Each element is made up of one or more segments
    that define drawing primitives.
'''

from __future__ import annotations
from typing import Sequence, Any, Union
import math

from .types import BBox, XY, Linestyle, Capstyle, Joinstyle, Align, Arcdirection, EndRef, RotationMode
from . import util
from .util import Point
from .backends import svg


def roundcorners(verts: Sequence[XY], radius: float=.5) -> Sequence[XY]:
    ''' Round the corners of polygon defined by verts.
        Works for convex polygons assuming radius fits inside.

        Args:
            verts: List of (x,y) pairs defining corners of polygon
            radius: Radius of curvature

        Adapted from:
        https://stackoverflow.com/questions/24771828/algorithm-for-creating-rounded-corners-in-a-polygon
    '''
    poly: list[Point] = []
    for v in range(len(verts))[::-1]:
        p1 = verts[v]
        p2 = verts[v-1]
        p3 = verts[v-2]

        dx1 = p2[0]-p1[0]
        dy1 = p2[1]-p1[1]
        dx2 = p2[0]-p3[0]
        dy2 = p2[1]-p3[1]

        angle = (math.atan2(dy1, dx1) - math.atan2(dy2, dx2))/2
        tan = abs(math.tan(angle))
        segment = radius/tan

        def getlength(x, y):
            return math.sqrt(x*x + y*y)

        def getproportionpoint(point, segment, length, dx, dy):
            factor = segment/length
            return point[0]-dx * factor, point[1]-dy * factor

        length1 = getlength(dx1, dy1)
        length2 = getlength(dx2, dy2)
        length = min(length1, length2)

        if segment > length:
            segment = length
            radius = length * tan

        p1cross = getproportionpoint(p2, segment, length1, dx1, dy1)
        p2cross = getproportionpoint(p2, segment, length2, dx2, dy2)

        dx = p2[0]*2 - p1cross[0] - p2cross[0]
        dy = p2[1]*2 - p1cross[1] - p2cross[1]
        L = getlength(dx, dy)
        d = getlength(segment, radius)
        circlepoint = getproportionpoint(p2, d, L, dx, dy)
        startangle = math.atan2(p1cross[1] - circlepoint[1], p1cross[0]-circlepoint[0])
        endangle = math.atan2(p2cross[1] - circlepoint[1], p2cross[0]-circlepoint[0])

        while endangle < startangle:
            endangle += 2*math.pi

        ph = util.linspace(startangle, endangle, 100)
        arc = [Point((circlepoint[0] + math.cos(i)*radius, circlepoint[1] + math.sin(i)*radius)) for i in ph]

        poly.extend(arc)
    poly.append(poly[0])  # Close the loop
    return poly


class Segment:
    ''' A segment path

        Args:
            path: List of (x,y) coordinates making the path
            color: Color for this segment
            lw: Line width for the segment
            ls: Line style for the segment '-', '--', ':', etc.
            capstyle: Capstyle for the segment: 'butt', 'round', 'square', ('projecting')
            joinstyle: Joinstyle for the segment: 'round', 'miter', or 'bevel'
            fill: Color to fill if path is closed
            zorder: Z-order for segment
    '''
    def __init__(self, path: Sequence[XY],
                 color: str=None,
                 lw: float=None,
                 ls: Linestyle=None,
                 capstyle: Capstyle=None,
                 joinstyle: Joinstyle=None,
                 fill: str=None,
                 zorder: int=None):
        self.path: Sequence[XY] = path  # Untranformed path
        self.zorder = zorder
        self.color = color
        self.fill = fill
        self.lw = lw
        self.ls = ls
        self.capstyle = capstyle
        self.joinstyle = joinstyle

    def xform(self, transform, **style) -> 'Segment':
        ''' Return a new Segment that has been transformed
            to its global position

            Args:
                transform: Transformation to apply
                style: Style parameters from Element to apply as default
        '''
        params: dict[str, Any] = {'zorder': self.zorder,
                                  'color': self.color,
                                  'fill': self.fill,
                                  'lw': self.lw,
                                  'ls': self.ls,
                                  'capstyle': self.capstyle,
                                  'joinstyle': self.joinstyle}
        style = {k: v for k, v in style.items() if k in params.keys()}
        params.update(style)
        return Segment(transform.transform_array(self.path), **params)

    def get_bbox(self) -> BBox:
        ''' Get bounding box (untransformed)

            Returns:
                Bounding box limits: (xmin, ymin, xmax, ymax)
        '''
        x = [p[0] for p in self.path]
        y = [p[1] for p in self.path]
        return BBox(min(x), min(y), max(x), max(y))

    def doreverse(self, centerx: float) -> None:
        ''' Reverse the path (flip horizontal about the center of the path) '''
        self.path = [util.mirrorx(p, centerx) for p in self.path[::-1]]

    def doflip(self) -> None:
        ''' Vertically flip the element '''
        self.path = [util.flip(p) for p in self.path]

    def draw(self, fig, transform, **style) -> None:
        ''' Draw the segment

            Args:
                fig: schemdraw.Figure to draw on
                transform: Transform to apply before drawing
                style: Default style parameters
        '''
        path = transform.transform_array(self.path)

        zorder = self.zorder if self.zorder is not None else style.get('zorder', 2)
        color = self.color if self.color else style.get('color', 'black')
        fill = self.fill if self.fill is not None else style.get('fill', None)
        ls = self.ls if self.ls else style.get('ls', '-')
        lw = self.lw if self.lw else style.get('lw', 2)
        capstyle = self.capstyle if self.capstyle else style.get('capstyle', 'round')
        joinstyle = self.joinstyle if self.joinstyle else style.get('joinstyle', 'round')

        if fill:  # Check if path is closed
            tlist = list(map(tuple, path))  # Need path as tuples for set()
            dofill = len(tlist) != len(set(tlist))  # Path has duplicates, can fill it
            if not dofill:
                fill = None
            elif fill is True:
                fill = color

        x = [p[0] for p in path]
        y = [p[1] for p in path]
        fig.plot(x, y, color=color, fill=fill,
                 ls=ls, lw=lw, capstyle=capstyle, joinstyle=joinstyle,
                 zorder=zorder)


class SegmentText:
    ''' A text drawing segment

        Args:
            pos: (x, y) coordinates for text
            label: Text to draw
            align: Tuple of (horizontal, vertical) alignment where horizontal
                is ('center', 'left', 'right') and vertical is ('center',
                'top', 'bottom')
            rotation: Rotation angle in degrees
            rotation_mode: See Matplotlib documentation. 'anchor' or 'default'.
            color: Color for this segment
            fontsize: Font size
            font: Font name/family
            zorder: Z-order for segment
    '''
    def __init__(self, pos: Sequence[float], label: str,
                 align: Align=None,
                 rotation: float=None,
                 rotation_mode: RotationMode=None,
                 color: str=None,
                 fontsize: float=14,
                 font: str=None,
                 zorder: int=None):
        self.xy = pos
        self.text = label
        self.align = align
        self.font = font
        self.fontsize = fontsize
        self.color = color
        self.rotation = rotation
        self.rotation_mode = rotation_mode
        self.zorder = zorder

    def doreverse(self, centerx: float) -> None:
        ''' Reverse the path (flip horizontal about the centerx point) '''
        self.xy = util.mirrorx(self.xy, centerx)

    def doflip(self) -> None:
        ''' Vertically flip the element '''
        self.xy = util.flip(self.xy)

    def xform(self, transform, **style) -> 'SegmentText':
        ''' Return a new Segment that has been transformed
            to its global position

            Args:
                transform: Transformation to apply
                style: Style parameters from Element to apply as default
        '''
        params: dict[str, Any] = {'align': self.align,
                                  'font': self.font,
                                  'fontsize': self.fontsize,
                                  'color': self.color,
                                  'rotation': self.rotation,
                                  'rotation_mode': self.rotation_mode,
                                  'zorder': self.zorder}

        style = {k: v for k, v in style.items() if k in params.keys()}
        params.update(style)
        return SegmentText(transform.transform(self.xy),
                           self.text, **params)

    def get_bbox(self) -> BBox:
        ''' Get bounding box (untransformed)

            Returns:
                Bounding box limits (xmin, ymin, xmax, ymax)
        '''
        w, h, _ = svg.text_size(self.text, 'Arial', self.fontsize)
        # h, w are in points, convert back to inches (/72)
        # then convert back to drawing units (*2)
        # This makes text the same point size regardless of inches_per_unit
        w = w/72*2
        h = h/72*2
        x = self.xy[0]
        y = self.xy[1]
        if self.align is not None:
            if self.align[0] == 'center':
                x -= w/2
            elif self.align[0] == 'right':
                x -= w
            if self.align[1] == 'center':
                y -= h/2
            elif self.align[1] == 'top':
                y -= h

        return BBox(x-.1, y-.1, x+w+.2, y+h+.2)

    def draw(self, fig, transform, **style) -> None:
        ''' Draw the segment

            Args:
                fig: schemdraw.Figure to draw on
                transform: Transform to apply before drawing
                style: Default style parameters
        '''
        xy = transform.transform(self.xy)
        color = self.color if self.color else style.get('color', 'black')
        fontsize = self.fontsize if self.fontsize else style.get('fontsize', style.get('size', 14))
        font = self.font if self.font else style.get('font', 'sans-serif')
        align = self.align if self.align else style.get('align', ('center', 'center'))
        rotation = self.rotation if self.rotation else style.get('rotation', 0)
        rotmode = self.rotation_mode if self.rotation_mode else style.get('rotation_mode', 'anchor')
        zorder = self.zorder if self.zorder is not None else style.get('zorder', 3)

        fig.text(self.text, xy[0], xy[1],
                 color=color, fontsize=fontsize, fontfamily=font,
                 rotation=rotation, rotation_mode=rotmode,
                 halign=align[0], valign=align[1], zorder=zorder)


class SegmentPoly:
    ''' A polygon segment

        Args:
            xy: List of (x,y) coordinates making the polygon
            closed: Draw a closed polygon (default True)
            cornerradius: Round the corners to this radius (0 for no rounding)
            color: Color for this segment
            lw: Line width for the segment
            ls: Line style for the segment
            fill: Color to fill if path is closed
            zorder: Z-order for segment
    '''
    def __init__(self, verts: Sequence[Sequence[float]],
                 closed: bool=True,
                 cornerradius: float=0,
                 color: str=None,
                 fill: str=None,
                 lw: float=None,
                 ls: Linestyle=None,
                 joinstyle: Joinstyle=None,
                 capstyle: Capstyle=None,
                 zorder: int=None):
        self.verts = verts
        self.closed = closed
        self.cornerradius = cornerradius
        self.color = color
        self.fill = fill
        self.joinstyle = joinstyle
        self.capstyle = capstyle
        self.zorder = zorder
        self.lw = lw
        self.ls = ls

    def doreverse(self, centerx: float) -> None:
        ''' Reverse the path (flip horizontal about the centerx point) '''
        self.verts = [util.mirrorx(v, centerx) for v in self.verts[::-1]]

    def doflip(self) -> None:
        ''' Vertically flip the element '''
        self.verts = [util.flip(p) for p in self.verts]

    def xform(self, transform, **style) -> 'SegmentPoly':
        ''' Return a new Segment that has been transformed
            to its global position

            Args:
                transform: Transformation to apply
                style: Style parameters from Element to apply as default
        '''
        params: dict[str, Any] = {'color': self.color,
                                  'fill': self.fill,
                                  'joinstyle': self.joinstyle,
                                  'capstyle': self.capstyle,
                                  'lw': self.lw,
                                  'ls': self.ls,
                                  'cornerradius': self.cornerradius,
                                  'zorder': self.zorder}
        style = {k: v for k, v in style.items() if k in params.keys()}
        params.update(style)
        return SegmentPoly(transform.transform_array(self.verts), **params)

    def get_bbox(self) -> BBox:
        ''' Get bounding box (untransformed)

            Returns:
                Bounding box limits (xmin, ymin, xmax, ymax)
        '''
        x = [p[0] for p in self.verts]
        y = [p[1] for p in self.verts]
        return BBox(min(x), min(y), max(x), max(y))

    def draw(self, fig, transform, **style) -> None:
        ''' Draw the segment

            Args:
                fig: schemdraw.Figure to draw on
                transform: Transform to apply before drawing
                style: Default style parameters
        '''
        fill = self.fill if self.fill is not None else style.get('fill', None)
        color = self.color if self.color else style.get('color', 'black')
        if fill is not None:
            if fill is True:
                fill = color
            if fill is False:
                fill = None
            elif fill == 'bg':
                fill = style.get('bgcolor', 'white')
        joinstyle = self.joinstyle if self.joinstyle else style.get('joinstyle', 'round')
        capstyle = self.capstyle if self.capstyle else style.get('capstyle', 'round')
        zorder = self.zorder if self.zorder is not None else style.get('zorder', 1)
        lw = self.lw if self.lw else style.get('lw', 2)
        ls = self.ls if self.ls else style.get('ls', '-')

        fill = color if fill is True else None if fill is False else fill
        verts = transform.transform_array(self.verts)

        if self.cornerradius > 0:
            verts = roundcorners(verts, self.cornerradius)

        fig.poly(verts, closed=self.closed, color=color, fill=fill, lw=lw, ls=ls,
                 capstyle=capstyle, joinstyle=joinstyle, zorder=zorder)


class SegmentCircle:
    ''' A circle drawing segment

        Args:
            center: (x, y) center of the circle
            radius: Radius of the circle
            color: Color for this segment
            lw: Line width for the segment
            ls: Line style for the segment
            fill: Color to fill if path is closed. True -> fill with element color.
            zorder: Z-order for segment
            ref: Flip reference ['start', 'end', None].
    '''
    def __init__(self, center: Sequence[float], radius: float,
                 color: str=None,
                 lw: float=None,
                 ls: Linestyle=None,
                 fill: bool | str | None=None,
                 zorder: int=None,
                 ref: EndRef=None):
        self.center = center
        self.radius = radius
        self.zorder = zorder
        self.color = color
        self.fill = fill
        self.lw = lw
        self.ls = ls

        # Reference for adding things AFTER lead extensions
        self.endref = ref

    def doreverse(self, centerx: float) -> None:
        ''' Reverse the path (flip horizontal about the centerx point) '''
        self.center = util.mirrorx(self.center, centerx)
        self.endref = {'start': 'end', 'end': 'start'}.get(self.endref)  # type: ignore

    def doflip(self) -> None:
        ''' Flip the segment up/down '''
        self.center = util.flip(self.center)

    def xform(self, transform, **style) -> 'SegmentCircle':
        ''' Return a new Segment that has been transformed
            to its global position

            Args:
                transform: Transformation to apply
                style: Style parameters from Element to apply as default
        '''
        params: dict[str, Any] = {'zorder': self.zorder,
                                  'color': self.color,
                                  'fill': self.fill,
                                  'lw': self.lw,
                                  'ls': self.ls,
                                  'ref': self.endref}
        style = {k: v for k, v in style.items() if k in params.keys()}
        params.update(style)
        return SegmentCircle(transform.transform(self.center),
                             self.radius, **params)

    def get_bbox(self) -> BBox:
        ''' Get bounding box (untransformed)

            Returns:
                Bounding box limits (xmin, ymin, xmax, ymax)
        '''
        xmin = self.center[0] - self.radius
        xmax = self.center[0] + self.radius
        ymin = self.center[1] - self.radius
        ymax = self.center[1] + self.radius
        return BBox(xmin, ymin, xmax, ymax)

    def draw(self, fig, transform, **style) -> None:
        ''' Draw the segment

            Args:
                fig: schemdraw.Figure to draw on
                transform: Transform to apply before drawing
                style: Default style parameters
        '''
        center = transform.transform(self.center)
        radius = transform.zoom * self.radius
        zorder = self.zorder if self.zorder is not None else style.get('zorder', 1)
        color = self.color if self.color else style.get('color', 'black')
        fill = self.fill if self.fill is not None else style.get('fill', None)
        ls = self.ls if self.ls else style.get('ls', '-')
        lw = self.lw if self.lw else style.get('lw', 2)

        if fill is not None:
            if fill is True:
                fill = color
            if fill is False:
                fill = None
            elif fill == 'bg':
                fill = style.get('bgcolor', 'white')
        
        fill = color if fill is True else None if fill is False else fill
        fig.circle(center, radius, color=color, fill=fill,
                   lw=lw, ls=ls, zorder=zorder)


class SegmentArrow:
    ''' An arrow drawing segment

        Args:
            tail: Start coordinate of arrow
            head: End coordinate of arrow
            headwidth: Width of arrowhead
            headlength: Length of arrowhead
            color: Color for this segment
            lw: Line width for the segment
            ls: Line style for the segment
            zorder: Z-order for segment
    '''
    def __init__(self, tail: Sequence[float], head: Sequence[float],
                 headwidth: float=None, headlength: float=None,
                 color: str=None, lw: float=None,
                 ref: EndRef=None,
                 zorder: int=None):
        self.tail = tail
        self.head = head
        self.zorder = zorder
        self.headwidth = headwidth
        self.headlength = headlength
        self.color = color
        self.lw = lw
        self.endref = ref

    def doreverse(self, centerx: float) -> None:
        ''' Reverse the path (flip horizontal about the centerx point) '''
        self.tail = util.mirrorx(self.tail, centerx)
        self.head = util.mirrorx(self.head, centerx)
        self.endref = {'start': 'end', 'end': 'start'}.get(self.endref)  # type: ignore

    def doflip(self) -> None:
        self.tail = util.flip(self.tail)
        self.head = util.flip(self.head)

    def xform(self, transform, **style) -> 'SegmentArrow':
        ''' Return a new Segment that has been transformed
            to its global position

            Args:
                transform: Transformation to apply
                style: Style parameters from Element to apply as default
        '''
        # See https://github.com/python/mypy/issues/5382
        # for this weird type annotation
        params: dict[str, Any] = {'zorder': self.zorder,
                                  'color': self.color,
                                  'lw': self.lw,
                                  'headwidth': self.headwidth,
                                  'headlength': self.headlength,
                                  'ref': self.endref}
        style = {k: v for k, v in style.items() if k in params.keys()}
        params.update(style)
        return SegmentArrow(transform.transform(self.tail),
                            transform.transform(self.head),
                            **params)

    def get_bbox(self) -> BBox:
        ''' Get bounding box (untransformed)

            Returns:
                Bounding box limits (xmin, ymin, xmax, ymax)
        '''
        hw = self.headwidth if self.headwidth else .1
        xmin = min(self.tail[0], self.head[0])
        ymin = min(self.tail[1], self.head[1])
        xmax = max(self.tail[0], self.head[0])
        ymax = max(self.tail[1], self.head[1])
        return BBox(xmin, ymin-hw, xmax, ymax+hw)

    def draw(self, fig, transform, **style) -> None:
        ''' Draw the segment

            Args:
                fig: schemdraw.Figure to draw on
                transform: Transform to apply before drawing
                style: Default style parameters
        '''
        tail = transform.transform(self.tail)
        head = transform.transform(self.head)
        zorder = self.zorder if self.zorder is not None else style.get('zorder', 1)
        color = self.color if self.color else style.get('color', 'black')
        lw = self.lw if self.lw else style.get('lw', 2)
        headwidth = self.headwidth if self.headwidth else style.get('headwidth', .2)
        headlength = self.headlength if self.headlength else style.get('headlength', .2)

        fig.arrow(tail[0], tail[1], head[0]-tail[0], head[1]-tail[1],
                  headwidth=headwidth, headlength=headlength,
                  color=color, lw=lw, zorder=zorder)


class SegmentArc:
    ''' An arc drawing segment

        Args:
            center: Center of the arc ellipse
            width: Width of the arc ellipse
            height: Height of the arc ellipse
            theta1: Starting angle in degrees
            theta2: Ending angle in degrees
            angle: Rotation of the ellipse defining the arc
            arrow: Direction of arrowhead ('cw' or 'ccw')
            color: Color for this segment
            lw: Line width for the segment
            ls: Line style for the segment
            fill: Color to fill if path is closed
            zorder: Z-order for segment
    '''
    def __init__(self, center: Sequence[float],
                 width: float, height: float,
                 theta1: float=35, theta2: float=-35,
                 arrow: Arcdirection=None,
                 angle: float=0,
                 color: str=None,
                 lw: float=None,
                 ls: Linestyle=None,
                 zorder: int=None):
        self.center = center
        self.width = width
        self.height = height
        self.theta1 = theta1
        self.theta2 = theta2
        self.arrow = arrow
        self.angle = angle
        self.color = color
        self.lw = lw
        self.ls = ls
        self.zorder = zorder

    def doreverse(self, centerx: float) -> None:
        ''' Reverse the path (flip horizontal about the centerx point) '''
        self.center = util.mirrorx(self.center, centerx)
        self.theta1, self.theta2 = 180-self.theta2, 180-self.theta1
        self.arrow = {'cw': 'ccw', 'ccw': 'cw'}.get(self.arrow, None)  # type: ignore

    def doflip(self) -> None:
        ''' Vertically flip the element '''
        self.center = util.flip(self.center)
        self.theta1, self.theta2 = -self.theta2, -self.theta1
        self.arrow = {'cw': 'ccw', 'ccw': 'cw'}.get(self.arrow, None)  # type: ignore

    def xform(self, transform, **style) -> 'SegmentArc':
        ''' Return a new Segment that has been transformed
            to its global position

            Args:
                transform: Transformation to apply
                style: Style parameters from Element to apply as default
        '''
        angle = self.angle + transform.theta
        params: dict[str, Any] = {'color': self.color,
                                  'lw': self.lw,
                                  'ls': self.ls,
                                  'zorder': self.zorder}
        style = {k: v for k, v in style.items() if k in params.keys()}
        params.update(style)
        return SegmentArc(transform.transform(self.center),
                          self.width*transform.zoom, self.height*transform.zoom, angle=angle,
                          theta1=self.theta1, theta2=self.theta2, **params)

    def get_bbox(self) -> BBox:
        ''' Get bounding box (untransformed)

            Returns:
                Bounding box limits (xmin, ymin, xmax, ymax)
        '''
        # Who wants to do trigonometry when we can just brute-force the bounding box?
        theta1, theta2 = math.radians(self.theta1), math.radians(self.theta2)
        # the phi parameter in parametric form is not the same as the angle along ellipse
        # (see https://www.petercollingridge.co.uk/tutorials/computational-geometry/finding-angle-around-ellipse/)
        t1 = math.atan2(self.width*math.sin(theta1), self.height*math.cos(theta1))
        t2 = math.atan2(self.width*math.sin(theta2), self.height*math.cos(theta2))
        while t2 < t1:
            t2 += 2*math.pi
        t = util.linspace(t1, t2, num=500)
        sint = list(map(math.sin, t))
        cost = list(map(math.cos, t))
        phi = math.radians(self.angle)
        cosphi = math.cos(phi)
        sinphi = math.sin(phi)
        rx = self.width/2
        ry = self.height/2
        xx = [self.center[0] + rx * ct*cosphi - ry * st*sinphi for st, ct in zip(sint, cost)]
        yy = [self.center[1] + rx * ct*sinphi + ry * st*cosphi for st, ct in zip(sint, cost)]
        return BBox(min(xx), min(yy), max(xx), max(yy))

    def draw(self, fig, transform, **style) -> None:
        ''' Draw the segment

            Args:
                fig: schemdraw.Figure to draw on
                transform: Transform to apply before drawing
                style: Default style parameters
        '''
        center = transform.transform(self.center)
        angle = self.angle + transform.theta

        zorder = self.zorder if self.zorder is not None else style.get('zorder', 1)
        color = self.color if self.color else style.get('color', 'black')
        ls = self.ls if self.ls else style.get('ls', '-')
        lw = self.lw if self.lw else style.get('lw', 2)

        width = self.width * transform.zoom
        height = self.height * transform.zoom
        fig.arc(center, width=width, height=height,
                theta1=self.theta1, theta2=self.theta2, angle=angle,
                color=color, lw=lw, ls=ls, zorder=zorder, arrow=self.arrow)


SegmentType = Union[Segment, SegmentText, SegmentPoly, SegmentArc,
                    SegmentCircle, SegmentArrow]